%% gh -- a compiler for Generic Haskell.
%% Copyright (c) 2001 -- 2005  The Generic Haskell Team. Utrecht University
%% 
%% This library is free software; you can redistribute it and/or modify
%% it under the terms of the GNU Lesser General Public License as
%% published by the Free Software Foundation; either version 2.1 of the
%% License, or (at your option) any later version.
%% 
%% This library is distributed in the hope that it will be useful, but
%% WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.
%% 
%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
%%
%% $Id: BuildSyntax.lhs 194 2005-01-12 09:16:08Z andres $
%%
%% author: Andres Loeh (andres@cs.uu.nl)

\section{Building syntax}

This file contains simple helper functions that allow to build
abstract syntax more easily. For the most part, the functions abbreviate
the rather verbose names of @AbstractSyntax@ (or @UHA_Syntax@), and they
pass the |unknownRange| argument automatically.

\begin{code}
module BuildSyntax (module BuildSyntax, module TypeConstruction) where

import AbstractSyntax
import UHA_Range
import Name
import TypeConstruction

simpleFunBinding                 :: Name -> Patterns -> Expression -> FunctionBinding
simpleFunBinding nm ps e         =  funBinding nm ps
                                      (RightHandSide_Expression
                                         unknownRange
                                         e
                                         MaybeDeclarations_Nothing
                                      )

funBinding                       :: Name -> Patterns -> RightHandSide -> FunctionBinding
funBinding nm ps rhs             =  FunctionBinding_FunctionBinding
                                      unknownRange
                                      (LeftHandSide_Function
                                         unknownRange
                                         nm ps
                                      )
                                      rhs

starK                            =  Kind_Star unknownRange
arrK                             =  Kind_Arrow unknownRange
varK                             =  Kind_Variable unknownRange
parK                             =  Kind_Parenthesized unknownRange


varX                             =  Export_Variable unknownRange
tycX                             =  Export_TypeOrClass unknownRange
tyccompleteX                     =  Export_TypeOrClassComplete unknownRange

varE                             =  Expression_Variable unknownRange
conE                             =  Expression_Constructor unknownRange
appE                             =  Expression_NormalApplication unknownRange
infE op l r                      =  Expression_InfixApplication unknownRange
                                      (MaybeExpression_Just l) 
                                      op
                                      (MaybeExpression_Just r)
parE                             =  Expression_Parenthesized unknownRange
litE                             =  Expression_Literal unknownRange
tappE                            =  Expression_TypeApplication unknownRange
lamE                             =  Expression_Lambda unknownRange

tupleE                           =  Expression_Tuple unknownRange

varP                             =  Pattern_Variable unknownRange
conP                             =  Pattern_Constructor unknownRange
infconP                          =  Pattern_InfixConstructor unknownRange
parP                             =  Pattern_Parenthesized unknownRange
litP                             =  Pattern_Literal unknownRange
wlcP                             =  Pattern_Wildcard unknownRange

stringL                          =  Literal_String unknownRange
intL                             =  Literal_Int unknownRange

tsigD                            =  Declaration_TypeSignature unknownRange
funbindD                         =  Declaration_FunctionBindings unknownRange
typeD                            =  Declaration_Type unknownRange
newtypeD                         =  Declaration_Newtype unknownRange

mkSimpleType                        :: Name -> Names -> SimpleType
mkSimpleType nm tvs                 =  SimpleType_SimpleType
				         unknownRange nm 
                                         (map (\x -> KindAnnotatedName_KindAnnotatedName
                                                       unknownRange x MaybeKind_Nothing
                                              ) tvs
                                         )

-- create a pattern of the form (ConDescr con _ _ _)

mkCondescrPat                    :: Name -> Pattern
mkCondescrPat con                =  parP
                                      $ conP condescrcondescrName
                                          [ litP (stringL $ extractName con)
                                          , wlcP
                                          , wlcP
                                          , wlcP
                                          , wlcP
                                          ]

mkLabdescrPat                    :: Name -> Pattern
mkLabdescrPat lab                =  parP
                                      $ conP labeldescrlabeldescrName
                                          [ litP (stringL $ extractName lab)
                                          , wlcP
                                          , wlcP
                                          ]


descrBinding                     :: Name -> Name -> Expression -> Declarations
descrBinding cl des exp          =  [Declaration_TypeSignature
                                       unknownRange
                                       [des]
                                       (conT cl)
                                    ,Declaration_FunctionBindings
                                       unknownRange
                                       [FunctionBinding_FunctionBinding
                                          unknownRange
                                          (LeftHandSide_Function
                                             unknownRange
                                             des
                                             []
                                          )
                                          (RightHandSide_Expression
                                             unknownRange
                                             exp
                                             MaybeDeclarations_Nothing
                                          )
                                       ]
                                    ]


mkCondescrDecl                   :: Name -> Name -> Int -> (Int,Int) -> Bool -> Declarations
mkCondescrDecl con tp arity pos l
                                 =  descrBinding 
                                      condescrName
                                      (cdName con)
                                      (mkCondescrExp con tp arity pos l)

mkCondescrExp                    :: Name -> Name -> Int -> (Int,Int) -> Bool -> Expression
mkCondescrExp con tp arity pos l =  parE
                                      $ appE (varE condescrcondescrName)
                                          [ litE (stringL $ extractName con)
                                          , litE (stringL $ extractName tp)
                                          , litE (intL $ show arity)
                                          , tupleE [litE (intL $ show $ fst pos),
                                                    litE (intL $ show $ snd pos)]
                                          , conE (hsBoolName l)
                                          , {- we have to define the fixity;
                                             - should become more sophisticated -}
                                            if is_symbol (extractName con) then
                                              parE $ appE (varE infixlName)
                                                       [litE (intL "9")]
                                            else varE nonfixName 
                                          ]
  where is_symbol (':':xs) = True
        is_symbol _        = False

mkLabdescrDecl                   :: Name -> Bool -> Name -> Declarations
mkLabdescrDecl tp s lab          =  descrBinding 
                                      labeldescrName
                                      (ldName lab)
                                      (mkLabdescrExp tp s lab)

mkLabdescrExp                    :: Name -> Bool -> Name -> Expression
mkLabdescrExp tp s lab           =  parE
                                      $ appE (varE labeldescrlabeldescrName)
                                          [ litE (stringL $ extractName lab)
                                          , litE (stringL $ extractName tp)
                                          , conE (hsBoolName s)
                                          ]

\end{code}
