%% gh -- a compiler for Generic Haskell.
%% Copyright (c) 2001 -- 2005  The Generic Haskell Team. Utrecht University
%% 
%% This library is free software; you can redistribute it and/or modify
%% it under the terms of the GNU Lesser General Public License as
%% published by the Free Software Foundation; either version 2.1 of the
%% License, or (at your option) any later version.
%% 
%% This library is distributed in the hope that it will be useful, but
%% WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.
%% 
%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
%%
%% $Id: BuildSyntax.lhs 138 2004-11-25 23:31:34Z andres $
%%
%% author: Andres Loeh (andres@cs.uu.nl)

%***************************************************************************
\section{Component utilities}
%***************************************************************************

This is the companion module for @Components.inc.lag@.

\begin{code}
module ComponentUtils where

import AbstractSyntax
import UHA_Range
import FMap
import TypeUtils
import KindUtils
import qualified Pretty as PP
import qualified Types
import qualified TypeIndexedTypes as TIT
import BuildSyntax
import Name
import ErrorUtils
import TypeReduce
\end{code}

%---------------------------------------------------------------------------
\subsection{Auxiliary attributes for structure types}
%---------------------------------------------------------------------------

\begin{code}

extraArguments                   :: TypeArgument -> [(Name,Name)] -> [Name]
extraArguments tp dependencies =
  case tp of
    TypeArgument_DependencyStyle _ _ d vs
                                 -> maybeNameToList d ++ [ specName as_f c
                                                         | c <- vs
                                                         , f <- dependencies
                                                         , let as_f = snd f
                                                         ]
    TypeArgument_Constructor _ n -> [n]
    TypeArgument_Label _ n       -> [n]
    _                            -> []
  where
    maybeNameToList MaybeName_Nothing  = []
    maybeNameToList (MaybeName_Just x) = [x]

prefixType                       :: TypeArgument -> (Type -> Type)
prefixType tp =
  case tp of
    TypeArgument_DependencyStyle _ n (MaybeName_Just _) _
      | n == conName             -> Types.funAT (Types.varT condescrName)
      | n == labelName           -> Types.funAT (Types.varT labeldescrName)
    TypeArgument_Constructor _ _ -> Types.funAT (Types.varT condescrName)
    TypeArgument_Label _ _       -> Types.funAT (Types.varT labeldescrName)
    _                            -> id

\end{code}

The following function, |kindOfTypeArgument|, tries to compute the kind
of a kind argument, given the environment of kinds associated with named
types. The only remarkable thing about the function is that for constructor
cases, we do not use a mapping from constructor names to datatype names
and lookup the kind of the datatype, but instead expect the constructor case
\emph{name} to be in the kind environment itself.

\begin{code}

kindOfTypeArgument               :: FMap Name Kind -> TypeArgument -> Kind
kindOfTypeArgument kinds typeArgument = 
  case typeArgument of
    TypeArgument_DependencyStyle _ n _ _
                                 -> applyFMM
                                      kinds
                                      noKindError
                                      n
    TypeArgument_NamedType _ n   -> applyFMM
                                      kinds
                                      noKindError
                                      n
    TypeArgument_Variable _ _    -> illegalTypeArgumentError
    TypeArgument_Constructor _ _ -> starToStarK
    TypeArgument_Label _ _       -> starToStarK
    TypeArgument_ConstructorCase _ n
                                 -> applyFMM
                                      kinds
                                      noKindError
                                    (concaseName n)
    TypeArgument_KindStarCase _  -> starK

 where
  starToStarK                  =  arrK starK starK
  noKindError n                =  internalError $
                                    "Components.kindOfTypeArgument: "
                                      ++ "no kind for "
                                      ++ PP.unPretty (PP.ppName n)
  illegalTypeArgumentError     =  internalError $
                                    "Components.kindOfTypeArgument: "
                                      ++ "illegal type argument (variable)"

\end{code}

%---------------------------------------------------------------------------
\subsection{Generating components}
%---------------------------------------------------------------------------

\begin{code}

makeComponent                    :: Kind            -- kind of the type
                                 -> KindIndexedType -- type of generic function
                                 -> Name            -- name of generic function
                                 -> TypeArgument    -- component
                                 -> Names           -- extra arguments
                                 -> FMap Name [(Name,Name)] -- dependencies
                                 -> TypeIndexedBindings
                                                    -- bindings of component
                                 -> Declarations    -- typesig plus final component
makeComponent kind kindIndexedType name typeArgument extraArguments
              dependencies bindings =
  let
    prefix                       =  prefixType typeArgument
    finalTypeArgument            =  case typeArgument of
                                      TypeArgument_KindStarCase _ ->
                                           varT (head varNames)
                                      _ -> typeArgumentToType typeArgument

    typesig                      =  Types.haskell98Type
                                 .  Types.normalize  -- possibly superfluous
                                 .  TIT.handleTypeIndexedTypes dependencies
                                 .  Types.normalize  -- must be prior to titypes because of ABS
                                 .  prefix
                                 .  forcePlainType "Components.makeComponent"
                                 $  Types.reduce
                                      (Types.applyKindIndexedType
                                         kindIndexedType kind
                                      )
                                      [finalTypeArgument]

    generatedName                =  componentName 
                                      name 
                                      (Types.typeArgumentName typeArgument)
  in
    [Declaration_TypeSignature
       unknownRange
       [generatedName]
       (Types.parenthesize typesig)
    ,Declaration_FunctionBindings
       unknownRange
       (map (transformBinding generatedName extraArguments) bindings)
    ]
 where
  transformBinding :: Name -> [Name] -> TypeIndexedBinding -> FunctionBinding
  transformBinding name extraArguments tib =
    case tib of
      TypeIndexedBinding_TypeIndexedBinding range _ _ patterns righthandside ->
        funBinding name (map varP extraArguments ++ patterns) righthandside
      TypeIndexedBinding_DefaultCase range n extended reassignments ->
        let  new       =  case reassignments of
                            MaybeReassignments_Nothing  -> 
                              makeRemap [Reassignment_Reassignment 
                                           unknownRange extended name]
                            MaybeReassignments_Just r   ->
                              makeRemap r
             (ps,e)    =  convert n extended typeArgument extraArguments kind new
        in   simpleFunBinding name ps e

  convert :: Name -> Name -> TypeArgument -> [Name] -> Kind -> (Name -> Name, Name -> Name) -> (Patterns,Expression)
  convert name extended ta extraArguments kind new = 
    let  kinds     =  splitKind kind
         vars      =  zip varNames kinds
         deps      =  map snd $ applyFMM dependencies (\n -> [(n,n)]) name
         edeps     =  map snd $ applyFMM dependencies (\n -> [(n,n)]) extended
         patterns  =  [varP (specName f t) | (t,_) <- vars, f <- deps]
         args      =  [mkLambda $ convert (apply new f) f (varTA t) [] k (invert new) 
                      | (t,k) <- vars, f <- deps]
         ename     =  specName extended (Types.typeArgumentName ta)
    in   (map varP extraArguments ++ patterns,appE (varE ename) (map (varE . makePlainName) extraArguments ++ args))

  apply         =  fst
  invert (x,y)  =  (y,x)

  mkLambda :: (Patterns,Expression) -> Expression
  mkLambda ([],e)  =  e
  mkLambda (ps,e)  =  parE (lamE ps e)

makeRemap    ::  Reassignments -> (Name -> Name, Name -> Name)
makeRemap x  =   (applyFM $ fromList new,applyFM $ fromList inv)
  where
    new  =  map single x
    inv  =  map (\(x,y) -> (y,x)) new
    single (Reassignment_Reassignment _ f t) = (f,t)

\end{code}

We have moved the ``expression replacement'' to the outside of |makeComponent|
now. In the earlier \GH\ versions there was an inconsistency here.

The function |convert| above corresponds closely to the function |convert| in
Chapter 14.4 of EGH on implementing Default Cases.

