%% gh -- a compiler for Generic Haskell.
%% Copyright (c) 2001 -- 2005  The Generic Haskell Team. Utrecht University
%% 
%% This library is free software; you can redistribute it and/or modify
%% it under the terms of the GNU Lesser General Public License as
%% published by the Free Software Foundation; either version 2.1 of the
%% License, or (at your option) any later version.
%% 
%% This library is distributed in the hope that it will be useful, but
%% WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.
%% 
%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
%%
%% $Id: ErrorMessages.lhs 194 2005-01-12 09:16:08Z andres $
%%
%% author: Andres Loeh (andres@cs.uu.nl)

%***************************************************************************
\section{Error messages}
%***************************************************************************

\begin{code}
module ErrorMessages where

import ErrorUtils
import AbstractSyntax
import PPrint
import Pretty (unPretty,ppName,ppEntity,ppKind,ppType)
import UHA_Range
\end{code}

Data type for error messages.

\begin{code}

data GHError = GHError_TypeMError  Name Range String Range String
             | GHError_TypeUError  Range Doc String
             | GHError_ConstrMError  Name Range Range
             | GHError_ConstrUError  Range Doc
             | GHError_ValueMError Name Range String Range String
             | GHError_ValueUError Range Doc String
             | GHError_TopLevelOnlyError Name Range String
             | GHError_ResolveError Range Name
             | GHError_NameClashError Range Name Names
             | GHError_DoubleDefinitionError Name Range Range
	     | GHError_OperatorAmbiguityError (Name,Int,Fixity) (Name,Int,Fixity)
             | GHError_PolyBindingMError Name Name Range Range
             | GHError_NoTypeSigError Name Range
             | GHError_NoPolyBindingWarn Name Range
             | GHError_PolyBindingWellError Name Range
             | GHError_PolyWellError Name Range
             | GHError_KindUAError Range Type [Type] Kind [Kind]
             | GHError_KindUTError Range Name String Kind Kind
             | GHError_KindUCError Range Name Names Kind
             | GHError_KindUTsError Range Name Kind
             | GHError_KindUGError Range Name Kind
             | GHError_KindUFError Range Kind
             | GHError_KindUnkIError Range Name
             | GHError_KindUnkTWarn  Range Name
             | GHError_KindIncGError Range Name (Int,Int) [Kind]
             | GHError_MPCTypeSigDepItemError Range Name
             | GHError_RequestAbstractError Name Type
             | GHError_ResolveDepError Range Name Name
             | GHError_RequestNonTerminate Int

isError :: GHError -> Bool
isError (GHError_NoPolyBindingWarn _ _) = False
isError (GHError_KindUnkTWarn _ _) = False
isError _ = True

printErrorV :: Int -> GHError -> Doc
printErrorV v e | isError e = printError e
                | v > 0     = printError e
                | otherwise = empty

printErrorsV :: Int -> [GHError] -> Doc
printErrorsV v [] = empty
printErrorsV v xs = empty <$> vsep (punctuate empty (map (printErrorV v) xs)) <$> empty <$> empty

printErrors :: [GHError] -> Doc
printErrors = printErrorsV 1

printError :: GHError -> Doc
printError (GHError_TypeMError name sl1 t1 sl2 t2)
  = errorLayout 
      [sl1,sl2]
      "error: name clash on type level"
      (paragraph $ 
         t1 ++ " and " ++ t2 ++ " at given locations both have name "
           ++ (unPretty $ ppName name)
      )

printError (GHError_TypeUError sl name t)
  = errorLayout 
      [sl]
      "error: unknown type identifier"
      (paragraph $
         unPretty name ++ " is not a known " ++ t
      )

printError (GHError_ConstrMError name sl1 sl2)
  = errorLayout 
      [sl1,sl2]
      "error: name clash for constructors"
      (paragraph $
         "Two constructors have name "
           ++ (unPretty $ ppName name)
      )


printError (GHError_ConstrUError sl name)
  = errorLayout 
      [sl]
      "error: unknown constructor"
      (paragraph $
         unPretty name ++ " is not a known constructors name"
      )

printError (GHError_ValueMError name sl1 t1 sl2 t2)
  = errorLayout 
      [sl1,sl2]
      "error: name clash on value level"
      (paragraph $
         t1 ++ " and " ++ t2 ++ " at given locations both have name "
           ++ (unPretty $ ppName name)
      )

printError (GHError_ValueUError sl name t)
  = errorLayout 
      [sl]
      "error: unknown type identifier"
      (paragraph $
         unPretty name ++ " is not a known " ++ t
      )

printError (GHError_TopLevelOnlyError name sl t)
  = errorLayout 
      [sl]
      "error: toplevel-only statement in local position"
      (paragraph $
         t ++ " " ++ (unPretty $ ppName name) ++ " has been declared locally, "
           ++ "but this is not (yet) allowed"
      )

printError (GHError_ResolveError r name)
  = errorLayout 
      [r]
      "error: unresolved name"
      (paragraph $
         (unPretty $ ppName name) ++ " is unknown"
      )

printError (GHError_NameClashError r name entities)
  = errorLayout
      [r]
      "error: name clash"
      (nest 4 $ ppName name <+> text "refers to the following entities:"
         <$> errorEntities entities
      )

printError (GHError_DoubleDefinitionError name r1 r2)
  = errorLayout
      [r1,r2]
      "error: multiple definitions"
      (paragraph $ 
         ("for " ++ (unPretty $ ppName name))
      )

printError (GHError_OperatorAmbiguityError (name1,prio1,assoc1) (name2,prio2,assoc2))
  = errorLayout
      [getRange name1,getRange name2]
      "error: amibuous use of infix operators"
      (ppName name1 <> text ", precedence" <+> int prio1 <> text "," <+> text (assoc assoc1)
        <$> ppName name2 <> text ", precedence" <+> int prio2 <> text "," <+> text (assoc assoc2)
      )
 where
  assoc (Fixity_Infixl _) = "left-associative"
  assoc (Fixity_Infixr _) = "right-associative"
  assoc (Fixity_Infix _)  = "non-associative"

printError (GHError_PolyBindingMError name st sl1 sl2)
  = errorLayout 
      [sl1,sl2]
      "error: multiply defined generic function binding"
      (paragraph $
         "for type " ++ (unPretty $ ppName st) ++ " and function "
            ++ (unPretty $ ppName name)
      )

printError (GHError_NoTypeSigError name sl)
  = errorLayout 
      [sl]
      "error: type signature missing"
      (paragraph $
         "for generic function " ++ (unPretty $ ppName name)
      )

printError (GHError_NoPolyBindingWarn name sl)
  = errorLayout 
      [sl]
      "warning: signature without definition"
      (paragraph $
         "for generic function " ++ (unPretty $ ppName name)
      )

printError (GHError_PolyBindingWellError name sl)
  = errorLayout 
      [sl]
      "error: malformed generic function"
      (paragraph $ 
         "generic function " ++ (unPretty $ ppName name)
           ++ " has a variable case at given location,\n"
           ++ "but has other cases."
      )

printError (GHError_PolyWellError name sl)
  = errorLayout
      [sl]
      "error: malformed kind-indexed type"
      (paragraph $ 
         "type " ++ (unPretty $ ppName name)
           ++ " needs exactly two cases, one for *, one for (->)"
      )

printError (GHError_KindUAError sl p1 p2 k1 k2)
  = errorLayout
      [sl]
      "error: kind unification failed"
      (text "in application of" 
         <$> indent 4 (ppType p1 <+> text "::" <+> ppKind k1)
         <$> text "to"
         <$> indent 4 (vsep (zipWith (\ t k -> ppType t <+> text "::" <+> ppKind k)
			       p2 k2
			    )
		      )
      )

printError (GHError_KindUTError sl name d ke ki)
  = errorLayout
      [sl]
      "error: kind unification failed"
      (text ("for " ++ d) <+> ppName name
         <$> text "expected kind" <+> ppKind ke
         <$> text "inferred kind" <+> ppKind ki
      )

printError (GHError_KindUCError sl name fname k)
  = errorLayout 
      [sl] 
      "error: constructor field does not have kind *"
      (text "inferred kind " <+> ppKind k 
         <$> indent 4 (text "for" <+> ff fname) 
         <$> indent 4 (text "in constructor" <+> ppName name)
      )
 where ff []  = text "unlabelled field"
       ff xs  = text "fields" <+> sep (map ppName xs)

printError (GHError_KindUTsError sl name k)
  = errorLayout
      [sl]
      "error: type synonym rhs does not have kind *"
      (text "inferred kind " <+> ppKind k
         <$> indent 4 (text "for rhs in declaration of type") <+> ppName name
      )

printError (GHError_KindUGError sl name k)
  = errorLayout
      [sl]
      "error: generic type signature does not have kind *"
      (text "inferred kind " <+> ppKind k
         <$> indent 4 (text "for the type signature of") <+> ppName name
      )

printError (GHError_KindUFError sl k)
  = errorLayout
      [sl]
      "error: universal quantification does not have kind *"
      (text "inferred kind " <+> ppKind k <+> text " instead"
      )

printError (GHError_KindUnkIError sl name)
  = errorLayout
      [sl]
      "error: unknown identifier during kind inference"
      (paragraph $
         "identifier " ++ unPretty (ppName name) ++ " not in scope"
      )

printError (GHError_KindUnkTWarn  sl name)
  = errorLayout 
      [sl]
      "warning: unknown type during kind infernce"
      (paragraph $
         "type " ++ unPretty (ppName name) ++ " unknown\n"
      )

printError (GHError_KindIncGError sl name (ag,ap) ks)
  = errorLayout
      [sl]
      "error: incompatible kinds for generic type variables in generic type signature"
      (text "generic function" <+> ppName name <+> text "of arity <" <> int ag <> text "|" <> int ap <> text ">"
         <$> text "has generic type variables of kinds" <+> hsep (punctuate (text ",") (map ppKind ks))
      )

printError (GHError_MPCTypeSigDepItemError sl name)
  = errorLayout
      [sl]
      "error: MPC-style generic type signature contains dependency items"
      (text "the type signature for" <+> ppName name <+> text "contains dependency items,"
         <$> text "but the kind variable in the type argument indicated an MPC-style definition."
         <$> text "Probable fix: remove the kind variable.")

printError (GHError_RequestAbstractError name tp)
  = errorLayout 
      []
      "error: specialisation request for abstract type"
      (text "generic function" <+> ppName name
         <$> text "needs to be specialised to" <+> ppType tp
         <$> (paragraph $ 
                "which is an abstract type for which there is no special case "
                  ++ "in the function definition."
             )
      )

printError (GHError_ResolveDepError sl name var)
  = errorLayout
      [sl]
      "error: unresolved dependency" 
      (text "for" <+> ppName name <+> text "on type variable" <+> ppName var <> text ".")

printError (GHError_RequestNonTerminate n)
  = errorLayout 
      []
      "error: computation of specialisation requests did not terminate"
      (paragraph $
         "after " ++ show n ++ " iterations.\n\n"
            ++ "You can try to increase the number of iterations with the\n"
            ++ "--cut command line option."
      )

errorEntities :: [Name] -> Doc
errorEntities = vcat . map errorEntity

errorEntity :: Name -> Doc
errorEntity = ppEntity 

\end{code}

