-- gh -- a compiler for Generic Haskell.
-- Copyright (c) 2001 -- 2005  The Generic Haskell Team. Utrecht University
-- 
-- This library is free software; you can redistribute it and/or modify
-- it under the terms of the GNU Lesser General Public License as
-- published by the Free Software Foundation; either version 2.1 of the
-- License, or (at your option) any later version.
-- 
-- This library is distributed in the hope that it will be useful, but
-- WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-- Lesser General Public License for more details.
-- 
-- You should have received a copy of the GNU Lesser General Public
-- License along with this library; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
--
-- $Id: FMap.hs 194 2005-01-12 09:16:08Z andres $
--
-- author: Andres Loeh (andres@cs.uu.nl)

{-
 - This is a wrapper for the hslibs FiniteMap module
 -}

module FMap
         (FMap,emptyFM,add,add',remove,find,applyFM,applyFM',applyFMM
         ,combine,modify,mapFM
         ,thenFM,mergeFM,domain,codomain,fromList,toList,showFM
         ,findL,addLFront,addLBack,addLUnion,addLReplace,removeL
         ,filterFM) where

import qualified FiniteMap
import List ((\\),union)
type FMap a b = FiniteMap.FiniteMap a b

emptyFM                        :: FMap a b
emptyFM                        =  FiniteMap.emptyFM

-- add throws away the previous

add                            :: Ord a => FMap a b -> (a,b) -> FMap a b
add fm (k,v)                   =  FiniteMap.addToFM fm k v

-- add' keeps the previous

add'                           :: Ord a => FMap a b -> (a,b) -> FMap a b
add' fm (k,v)                  =  case find fm k of
                                    Nothing -> add fm (k,v)
                                    Just _  -> fm

remove                         :: Ord a => FMap a b -> a -> FMap a b
remove fm v                    =  FiniteMap.delFromFM fm v

find                           :: Ord a => FMap a b -> a -> Maybe b
find fm v                      =  FiniteMap.lookupFM fm v

applyFM                        :: Ord a => FMap a a -> a -> a
applyFM f a                    =  case find f a of
                                    Nothing -> a
                                    Just x  -> x

applyFM'                       :: Ord a => FMap a b -> a -> b
applyFM' f a                   =  case find f a of
                                    Nothing -> error "FMap.applyFM': finite map undefined"
                                    Just x  -> x

applyFMM                       :: Ord a => FMap a b -> (a -> b) -> a -> b
applyFMM f m a                 =  case find f a of
                                    Nothing -> m a
                                    Just x  -> x

combine                        :: Ord a => FMap a a -> FMap a a -> FMap a a
combine mg mf                  =  FiniteMap.plusFM_C 
                                           (curry fst)
                                           (FiniteMap.mapFM (const $ applyFM mg) mf)
                                           mg

modify                         :: Ord a => FMap a b -> a -> (b -> b) -> FMap a b
modify fm k f                  =  {- FiniteMap.addToFM_C (\o _ -> f o) fm k (error "FMap.modify") -}
                                  case find fm k of
                                    Nothing  -> fm
                                    Just old -> add fm (k,f old)

mapFM                          :: Ord a => (b -> c) -> FMap a b -> FMap a c
mapFM f fm                     =  FiniteMap.mapFM (const f) fm

thenFM                         :: Ord a => FMap a b -> FMap a b
                               -> (FMap a b -> b -> b)
                               -> FMap a b
thenFM   mg mf a               =  FiniteMap.mapFM (const $ a mf) mg

-- filters

filterFM                       :: Ord a => (a -> b -> Bool) -> FMap a b -> FMap a b
filterFM                       =  FiniteMap.filterFM

-- merges two finite maps; the left shadows the right

mergeFM                        :: Ord a => FMap a b -> FMap a b -> FMap a b
mergeFM mg mf                  =  FiniteMap.plusFM mf mg

-- list finite map functions

findL                          :: Ord a => FMap a [b] -> a -> [b]
findL fm k                     =  applyFMM fm (const []) k

addLFront                      :: Ord a => FMap a [b] -> (a,[b]) -> FMap a [b]
addLFront fm (k,v)             =  let old = applyFMM fm (const []) k
                                      new = v ++ old
                                  in  add fm (k,new)

addLBack                       :: Ord a => FMap a [b] -> (a,[b]) -> FMap a [b]
addLBack  fm (k,v)             =  let old = applyFMM fm (const []) k
                                      new = old ++ v
                                  in  add fm (k,new)

addLUnion                      :: (Ord a,Eq b) => FMap a [b] -> (a,[b]) -> FMap a [b]
addLUnion fm (k,v)             =  let old = applyFMM fm (const []) k
                                      new = old `union` v
                                  in  add fm (k,new)

addLReplace                    :: Ord a => FMap a [b] -> (a,[b]) -> FMap a [b]
addLReplace fm (k,v)           =  let old = applyFMM fm (const []) k
                                      new = v
                                  in  add fm (k,new)

removeL                        :: (Ord a,Eq b) => FMap a [b] -> (a,[b]) -> FMap a [b]
removeL   fm (k,v)             =  modify fm k (\\v)

domain                         :: FMap a b -> [a]
domain                         =  FiniteMap.keysFM

codomain                       :: FMap a b -> [b]
codomain                       =  FiniteMap.eltsFM

fromList                       :: Ord a => [(a,b)] -> FMap a b
fromList                       =  FiniteMap.listToFM

toList                         :: FMap a b -> [(a,b)]
toList                         =  FiniteMap.fmToList

showFM                         :: (Show a,Show b) => FMap a b -> String
showFM f                       = unlines
                               . map (\(x,y) -> show x ++ " |-> " ++ show y)
                               . toList 
                               $ f
