%% gh -- a compiler for Generic Haskell.
%% Copyright (c) 2001 -- 2005  The Generic Haskell Team. Utrecht University
%% 
%% This library is free software; you can redistribute it and/or modify
%% it under the terms of the GNU Lesser General Public License as
%% published by the Free Software Foundation; either version 2.1 of the
%% License, or (at your option) any later version.
%% 
%% This library is distributed in the hope that it will be useful, but
%% WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.
%% 
%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
%%
%% $Id: FileNameUtils.lhs 194 2005-01-12 09:16:08Z andres $
%% 
%% author: Jan de Wit (jwit@cs.uu.nl), Andres Loeh (andres@cs.uu.nl)

> module FileNameUtils          ( extension
>			        , filename
>			        , basename
>                               , dirname
>                               , isAbsolute
>                               , expandPath
>                               , chaseFile
>                               , descendFrom
>                               , Call
>                               , checkCompilation
>			        ) where
>
> import IO
> import System
> import Directory
> import List
> import Monad (filterM)
> import OSSpecific

> type FileName                 =  String

> extension                     :: FileName -> Maybe String
> extension fn                  =  f False [] fn 
>   where
>   f found acc [] | found      =  Just (reverse acc)
>                  | not found  =  Nothing 
>   f found acc ('.':cs)        =  f True  []      cs
>   f found acc (c  :cs)        =  f found (c:acc) cs

> dirname                       :: FileName -> String
> dirname fn                    =  f [] [] fn
>   where
>   f res acc []                =  reverse res
>   f res acc (c:cs) 
>     | c `elem` directorySeparators 
>                               =  f (c : acc ++ res) [] cs
>     | otherwise               =  f res       (c : acc) cs

> isAbsolute                    :: FileName -> Bool
> isAbsolute fn                 =  let nodrive = snd (splitDrive fn)
>                                  in  not (null nodrive) 
>                                      && head nodrive `elem` directorySeparators

> filename                      :: FileName -> String
> filename fn                   =  f [] fn 
>   where
>   f acc []                    =  reverse acc 
>   f acc (c:cs)
>     | c `elem` directorySeparators
>                               =  f []      cs
>     | otherwise               =  f (c:acc) cs

> basename                      :: FileName -> String
> basename fn                   =  takeWhile (/= '.') (filename fn)  

|expandPath| does two things: it replaces curly braced strings with
environment entries, if present; furthermore, if the path ends with
more than one directory separator, all subpaths are added ...

> expandPath                    :: [String] -> IO [String]
> expandPath s                  =  do let s' = concatMap (splitOn environmentSeparators) s
>                                     s'' <- mapM expandEnvironment s'
>                                     s''' <- mapM findSubPaths (concat s'')
>                                     return (nub $ concat s''')

> findSubPaths                  :: String -> IO [String]
> findSubPaths ""               =  return []
> findSubPaths s                =  let rs = reverse s
>                                      (sep,rs') = span (`elem` directorySeparators) rs
>                                      s'   = reverse rs'
>                                      sep' = reverse sep
>                                  in  if   null s' 
>                                      then return [[head sep']] {- we don't descend from root -}
>                                      else if   length sep < 2
>                                           then return [s]
>                                           else descendFrom s'

> descendFrom                   :: String -> IO [String]
> descendFrom s                 =  catch (do  d <- getDirectoryContents s
>                                             {- no hidden files, no parents -}
>                                             let d' = map (\x -> s ++ [directorySeparator] ++ x)
>                                                    . filter ((/='.') . head) . filter (not . null) $ d
>                                             d'' <- filterM doesDirectoryExist d'
>                                             d''' <- mapM descendFrom d''
>                                             return (s : concat d''')
>                                        )
>                                        (const $ return [])

> expandEnvironment             :: String -> IO [String]
> expandEnvironment s           =  case break (=='{') s of
>                                    (s',"")    -> return [s]
>                                    (s','{':r) -> case break (=='}') r of
>                                                    (e,"") -> return [s]
>                                                    (e,'}':r') -> findEnvironment e s' r'
>   where findEnvironment       :: String -> String -> String -> IO [String]
>         findEnvironment e a o =  do er <- try (getEnv e)
>                                     return $ either (const [])
>                                                     (map (\x -> a ++ x ++ o) . splitOn environmentSeparators)
>                                                     er

> splitOn                       :: String -> String -> [String]
> splitOn b s                   =  case dropWhile (`elem` b) s of
>                                    "" -> []
>                                    s' -> w : splitOn b s''
>                                            where (w,s'') = break (`elem` b) s'

> chaseFile                     :: [String]    {- search path -}
>                               -> FileName -> IO (String,FileName)
> chaseFile p fn | isAbsolute fn
>                               =  t fn
>                | p == []      =  chaseFile ["."] fn
>                | otherwise    =  s $ map (\d -> t (md d ++ fn)) p'
>   where
>   md cs | last cs `elem` directorySeparators
>                               =  cs
>         | otherwise           =  cs ++ [directorySeparator]
>   t f                         =  catch (readFile f >>= \x -> return (x,f))
>                                        (\_ -> ioError $ userError $ "File `" ++ fn ++ "' not found.\n")
>   s []                        =  ioError 
>                               $  userError $ "File `" ++ fn ++ "' not found in search path:\n" ++ showpath
>   s (x:xs)                    =  catch x (\_ -> s xs)
>   p' | "." `elem` p           = p
>      | otherwise              = "." : p
>   showpath                    =  concatMap (\x -> "   " ++ x ++ "\n") p

> type Call                     =  (String,FileName,FileName,FileName)

> checkCompilation              :: [String]    {- search path -}
>                               -> String      {- module name -}
>                               -> IO (Bool,Maybe Call) {- required?, full call-}
> checkCompilation sp m         =  do let ghi_file = m ++ ".ghi"
>                                         ghs_file = m ++ ".ghs"
>                                         hs_file  = m ++ ".hs"
>                                     (ghi_ex,ghi_pos) <- chase ghi_file
>                                     (ghs_ex,ghs_pos) <- if   ghi_ex
>                                                         then let ghs = dirname ghi_pos ++ ghs_file
>                                                              in  doesFileExist ghs >>= \x -> return (x,ghs)
>                                                         else chase ghs_file
>                                     (hs_ex,hs_pos)   <- if   ghi_ex
>                                                         then let hs = dirname ghi_pos ++ hs_file
>                                                              in  doesFileExist hs >>= \x -> return (x,hs)
>                                                         else chase hs_file
>                                     if   not ghi_ex
>                                      then if   ghs_ex
>                                           then return (True,Just (m,ghs_pos,dirname ghs_pos ++ ghi_file,dirname ghs_pos ++ hs_file))
>                                           else return (False,Nothing)
>                                             {- insert Haskell compilation here -}
>                                      else if   ghs_ex
>                                           then do r <- catch ( do ghi_t <- getModificationTime ghi_pos
>                                                                   ghs_t <- getModificationTime ghs_pos
>                                                                   return (ghi_t < ghs_t)
>                                                              ) (\_ -> return False)
>                                                   return (r,Just (m,ghs_pos,ghi_pos,dirname ghs_pos ++ hs_file))
>                                           else return (False,Nothing) {- insert Haskell compilation here -}
>   where chase f               =  catch (chaseFile sp f >>= \(_,x) -> return (True,x)) 
>                                        (\_ -> return (False,f))
