%% gh -- a compiler for Generic Haskell.
%% Copyright (c) 2001 -- 2005  The Generic Haskell Team. Utrecht University
%% 
%% This library is free software; you can redistribute it and/or modify
%% it under the terms of the GNU Lesser General Public License as
%% published by the Free Software Foundation; either version 2.1 of the
%% License, or (at your option) any later version.
%% 
%% This library is distributed in the hope that it will be useful, but
%% WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.
%% 
%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
%%
%% $Id: KindUtils.lhs 194 2005-01-12 09:16:08Z andres $
%%
%% author: Andres Loeh (andres@cs.uu.nl)

%***************************************************************************
\section{Kind inference utilities}
%***************************************************************************

\begin{code}
module KindUtils where

import FMap
import AbstractSyntax
import Maybe
import UHA_Range
import Name
import BuildSyntax
\end{code}

%---------------------------------------------------------------------------
\subsection{Some simple functions}
%---------------------------------------------------------------------------

Create a simple chain of Star applications.

\begin{code}
simpleKind                       :: Int -> Kind
simpleKind 0                     =  starK
simpleKind n                     =  arrK starK (simpleKind (n-1))
\end{code}

Compute the rank of a kind. 

\begin{code}
rank                             :: Kind -> Int
rank (Kind_Star _)               =  1
rank (Kind_Arrow _ l r)          =  max (1 + rank l) (rank r)
rank (Kind_Parenthesized _ k)    =  rank k
\end{code}

Determine the number of arguments of a kind.

\begin{code}
kindargs                         :: Kind -> Int
kindargs (Kind_Star _)           =  0
kindargs (Kind_Arrow _ l r)      =  1 + kindargs r
kindargs (Kind_Parenthesized _ k)=  kindargs k
\end{code}

Check whether a given kind is a kind variable. A generic function
is called \emph{restricted} iff the kind of its type variable is
not a kind variable.

\begin{code}
isKindVar                        :: Kind -> Bool
isKindVar (Kind_Variable _ _)    =  True
isKindVar _                      =  False

isRestricted                     :: Kind -> Bool
isRestricted                     =  not . isKindVar
\end{code}

Finally, we need the function to remove all remaining variables from 
a kind. This is achieved by transforming a given substitution.

\begin{code}
removeKindVars                   :: Kind -> KSubst -> KSubst
removeKindVars k s               =  rkv (applyKind s k) s
  where rkv (Kind_Star _)      s =  s
        rkv (Kind_Arrow _ l r) s =  rkv r (rkv l s)
        rkv (Kind_Variable _ v) s=  add s (v,starK)
	rkv (Kind_Parenthesized _ k) s
	                         =  rkv k s
        rkv Kind_Error         s =  s
\end{code}

%---------------------------------------------------------------------------
\subsection{Kind unification}
%---------------------------------------------------------------------------

Unification of two kinds is always relative to a current substitution.
The substitution is changed so that applied to both given kinds it will
result in their most general unifier. If that is not possible, |Nothing|
is returned.

\begin{code}
type KSubst                      =  FMap Name Kind

unify                            :: Kind -> Kind  -- two input kinds
                                 -> KSubst -> (KSubst,Bool)
unify k1 k2 s                    =  unify' (applyKind s k1) (applyKind s k2) s
  where
\end{code}

If one kind is an error kind, then unification succeeds without changes.

\begin{code}
  unify' Kind_Error _        s   =  (s,True)
  unify' _        Kind_Error s   =  (s,True)
\end{code}

If both kinds are |Kind_Star|, then we succeed without doing something.

\begin{code}
  unify' (Kind_Star _) (Kind_Star _) s  
                                 =  (s,True)
\end{code}

If one of both kinds is a variable, then we bind the variable to
the other kind.

\begin{code}
  unify' (Kind_Variable _ v) k s =  varBind v k s
  unify' k (Kind_Variable _ v) s =  varBind v k s
\end{code}

In the case of two arrow kinds we first unify the left hand sides,
then unify the right hand sides making use of the knowledge we have
gained from the unification of the left hand sides. In the end, the
composition of both substitutions is returned.

\begin{code}
  unify' (Kind_Arrow _ l1 r1)
         (Kind_Arrow _ l2 r2) s  =  let (s1,p ) = unify l1 l2 s
                                        (s2,p') = unify r1 r2 s1
                                    in  (s2,p && p')
\end{code}

All other kinds cannot be unified.

\begin{code}
  unify' _ _ s                   =  (s,False)  -- !!
\end{code}

A simple version of |unify| just checks whether two kinds could be
unified.

\begin{code}
unifyable                        :: Kind -> Kind -> Bool
unifyable k1 k2                  =  snd $ unify k1 k2 emptyFM
\end{code}

Binding a variable to a kind in a substitution.

\begin{code}
varBind                          :: Name -> Kind -> KSubst -> (KSubst,Bool)
varBind v k@(Kind_Variable _ v') s
  | v == v'                      =  (s,True)   -- nothing to do
  | otherwise                    =  (add s (v,k),True)
varBind v k s
  | v `elemKV` k                 =  (s,False)  -- this is not good
  | otherwise                    =  (add s (v,k),True)
\end{code}

The function elemKV checks for occurences of a kind variable in a
kind.

\begin{code}
elemKV                           :: Name -> Kind -> Bool
elemKV v (Kind_Arrow _ l r)      =  elemKV v l || elemKV v r
elemKV v (Kind_Star _)           =  False
elemKV v (Kind_Variable _ v')    =  v == v'
elemKV v (Kind_Parenthesized _ k)=  elemKV v k
elemKV v (Kind_Error)            =  False
\end{code}

Apply a substitution to a kind.

\begin{code}
applyKind                        :: KSubst -> Kind -> Kind
applyKind subst k@(Kind_Star _)  =  k
applyKind subst (Kind_Arrow _ l r) 
                                 =  arrK (applyKind subst l) 
				         (applyKind subst r)
applyKind subst k@(Kind_Variable _ v)    
                                 =  case find subst v of
                                      Just x -> applyKind subst x
                                      _      -> k
applyKind subst (Kind_Parenthesized _ k)      
                                 =  parK (applyKind subst k)
applyKind subst k@(Kind_Error)   =  k
\end{code}

Do the same, but without recursive applications.

\begin{code}
substKind                        :: KSubst -> Kind -> Kind
substKind subst k@(Kind_Star _)  =  k
substKind subst (Kind_Arrow _ l r) 
                                 =  arrK (substKind subst l)
                                         (substKind subst r)
substKind subst k@(Kind_Variable _ v)    
                                 =  case find subst v of
                                      Just x -> x
                                      _      -> k
substKind subst (Kind_Parenthesized _ k)
                                 =  parK (substKind subst k)
substKind subst k@(Kind_Error)   =  k
\end{code}

Compose two kind substitutions.

\begin{code}
compKind                        :: KSubst -> KSubst -> KSubst
compKind sub1 sub2              =  (sub1 `thenFM` sub2) applyKind
\end{code}

Split a kind. Returns all arguments in a list.
For instance, |(* -> *) -> * -> *| would be split
into |[* -> *, *]|.

\begin{code}

splitKind                           ::  Kind -> [Kind]
splitKind (Kind_Parenthesized _ k)  =   splitKind k
splitKind (Kind_Arrow _ l r)        =   l : splitKind r
splitKind _                         =   []

\end{code}

This variant of |splitKind| strips a specific number of
arguments from a kind.

\begin{code}

splitKindArgs :: Int -> Kind -> ([Kind],Kind)
splitKindArgs n k = splitKindArgs' n [] k

splitKindArgs' 0 acc k                         =  (reverse acc,k)
splitKindArgs' n acc (Kind_Parenthesized _ k)  =  splitKindArgs' n acc k
splitKindArgs' n acc (Kind_Arrow _ l r)        =  splitKindArgs' (n-1) (l:acc) r
splitKindArgs' n acc _                         =  splitKindArgs' (n-1) (e:acc) e
  where e = Kind_Error

\end{code}

The function |getKind| tries, given a kind environment, to
determine the kind of a type term. It returns an error kind
if it can't.

\begin{code}

getKind :: KSubst -> Type -> Kind
getKind kinds t = case t of
  Type_Application _ _ f xs  ->
    let  kf        =  getKind kinds f
         (kxs,kr)  =  splitKindArgs (length xs) kf
    in   kr
  Type_Variable _ n        ->  applyFMM kinds (const Kind_Error) n
  Type_Constructor _ n     ->  applyFMM kinds (const Kind_Error) n
  Type_Qualified _ _ _     ->  starK
  Type_Forall _ _ _        ->  starK
  Type_Exists _ _ _        ->  starK
  Type_Parenthesized _ t   ->  getKind kinds t
  Type_TypeApplication _ _ _ -> 
    Kind_Error {- we'd have to apply a kind-indexed kind -}
  Type_KindApplication _ _ _ ->
    Kind_Error {- shouldn't occur for base types anyway -}
  _ ->
    Kind_Error {- the other cases shouldn't occur -}

\end{code}

The function |orKind| is a monadic plus on kinds.

\begin{code}

orKind :: Kind -> Kind -> Kind
orKind Kind_Error  k  =  k
orKind k           _  =  k

\end{code}
