%% gh -- a compiler for Generic Haskell.
%% Copyright (c) 2001 -- 2005  The Generic Haskell Team. Utrecht University
%% 
%% This library is free software; you can redistribute it and/or modify
%% it under the terms of the GNU Lesser General Public License as
%% published by the Free Software Foundation; either version 2.1 of the
%% License, or (at your option) any later version.
%% 
%% This library is distributed in the hope that it will be useful, but
%% WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.
%% 
%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
%%
%% $Id: Main.lhs 205 2005-01-13 12:27:46Z andres $
%%
%% author: andres@cs.uu.nl

%***************************************************************************
\section{Acknowledgements}
%***************************************************************************

Jan de Wit has written a prototype of the Generic Haskell compiler.

%***************************************************************************
\section{Main program}
%***************************************************************************

%---------------------------------------------------------------------------
\subsection{About the structure}
%---------------------------------------------------------------------------

Currently, the Generic Haskell compiler consists of native Haskell
modules as well as attribute grammar (\AG) files and grammar
descriptions for the frown parser generator. 

You can identify the type of the file by its extension:

\begin{center}
\begin{tabular}{ll} \hline 
|.hs|  & Haskell \\
|.lhs| & literate Haskell \\
|.lag| & literate \AG \\
|.inc.lag| & literate \AG include \\
|.lg|  & literate frown grammar \\ \hline
\end{tabular}
\end{center}

\AG source files are feeded into the |UU_AG| system to produce Haskell
files (which have the same name, but the extension |.hs|). \AG include
files are not standalone \AG source files, but are included into other
\AG files. Thus, there are no Haskell files of the same name.

Frown source files contain grammar descriptions that are used to
generate a parser. The generated Haskell file has again the same name
as the grammar file, but the extension |.hs|.

The file |Main.lhs| is the central Haskell module. It handles
the processing of command line options and the calls different
phases of the compiler. 

These are:
\begin{itemize}
\item parsing of the input file (|GHParser_NG|),
\item handling of interface files and imports (|Imports|),
\item static error analysis (|Analysis|),
\item compiler and output code generator (|GH|).
\end{itemize}

\paragraph{Compiling the program} Please refer to the User's
Guide and the distribution installation instructions for 
information on how to build and compile the program.

\paragraph{Running in hugs/ghci} If you want to run Generic
Haskell out of an interactive environment, then |findAndCompile|
is the function to call from the interactive command line.

\paragraph{More information} For more information about how
to use the \GH\ compiler, please look at the \GH User's Guide.

\begin{code}

module Main where

\end{code}

%---------------------------------------------------------------------------
\subsection{Auxiliary Haskell modules}
%---------------------------------------------------------------------------

Naturally, we need to import a lot of things. These are either
standard libraries or modules that ship with \GH.

\begin{code}
import ATermLib
import System                   -- these are for option handling
import System.Console.GetOpt    -- etc.
import Maybe
import List hiding (find)
import Monad
import IO
{-
-- import qualified ModuleLoader 
-- import qualified Optimizer
-- import qualified TestOptimizer
-}

-- import IOExts                   {- for debugging -}
\end{code}

Then there are a few basic utility libraries that are part of
Generic Haskell (or are, at least, distributed with Generic
Haskell).

\begin{code}
import FileNameUtils            {- basic file handling -}
import Name                     {- names of generated functions -}
import Utils                    {- general utilities -}
import PPrint (displayS,renderPretty,displayIO)
import qualified PPrint
-- import Pretty                   {- wrapper around pretty printing library -}
import Configure
import OSSpecific
\end{code}

Sometimes, the need arises to show debugging output even in
|Main|. In these cases, it comes handy to import the
module that pretty prints the abstract syntax of Generic
Haskell.

\begin{code}
-- temporary!!
-- import PrettyPrint as PP
\end{code}

%---------------------------------------------------------------------------
\subsection{Compiler phases}
%---------------------------------------------------------------------------

We need the abstract grammar with all extensions, and the parser.

\begin{code}
import AbstractSyntax
-- import AbstractPP
import qualified UHA_Parser as P
import qualified Pretty as PP
\end{code}

We need |Imports.hs| to deal with imported modules, 
|NameReplace.hs| to resolve all names,
|Analysis.hs| to do static error analysis.
We then import |GH.hs| that is supposed to deal with almost everything.

\begin{code}
import qualified Imports as I
import qualified NameReplace as NR
import qualified ResolveOperators as RO
-- import qualified InitialTypes as IT
-- import qualified StaticAnalysis as SA
-- begin StaticAnalysis related modules
-- import qualified MyAssocList as MAL
-- import qualified Strategy as S
-- import qualified SAMessages as SAM
-- end 
-- import qualified Analysis as A
import qualified GH as GH
\end{code}

%---------------------------------------------------------------------------
\subsection{Compilation}
%---------------------------------------------------------------------------

The function |findAndCompile| determines the correct input and
output file name for the call to the compiler. The extension defaults
to |ghs|, but can be different. The output file is automatically the
same basename with extension |hs|. The list of flags contains the
command line options. The corresponding data type is described near
the end of this file.

\begin{code}

findAndCompile                  :: Flags -> String -> IO ()
findAndCompile opts name        =  
  let make                      =  case fCallMode opts of { Make -> True; _ -> False }
      dir'                      =  dirname name
      dir                       =  if null dir' then "." else dir'
      file                      =  filename name
      odir                      =  case fOutput opts of { Nothing -> dir; Just d -> d }
      continue                  =  if make then
                                     (\c -> determineDependencies 
                                              True opts [] c 
                                            >> return ()
                                     )
                                   else 
                                     (\(b,g,i,r) -> compile opts b g i r)
  in  case (extension file) of 
        Nothing                 -> let obase = odir ++ [directorySeparator] ++ file
                                   in  continue (obase
                                                ,name ++ ".ghs"
                                                ,obase ++ ".ghi" 
                                                ,obase ++  ".hs")
        (Just "hs")             -> error "Source file is Haskell file."
        (Just ext)              -> 
          let bname = take (length file - (length ext + 1)) file
              obase = odir ++ [directorySeparator] ++ bname
          in  continue (obase,name
                       ,obase ++ ".ghi"
                       ,obase ++  ".hs"
                       )

\end{code}

The function |determineDependencies| issues the correct sequence
of |compile| commands to compile a file with dependencies.

\begin{code}

determineDependencies           :: Bool -> Flags -> [Call] -> Call -> IO Bool
determineDependencies top opts known from@(frommod,fromghs,fromghi,fromhs)
                                =  do
    -- copied from compile
    let verbosity               =  fVerbose opts
    sp                          <- if top then
                                     expandPath $ reverse (fLibPath opts) ++ searchPath
                                   else
                                     return []
    let atVerbosity n s         =  when (verbosity >= n) $ putStrLn s 

    -- Parse input file
    (pM,fn)                     <- (if top then id else (\a -> catch a (\ _ -> return (Nothing,fromghs))))
                                     $ P.parseFile verbosity sp fromghs
                                        
    unless (null fn) (atVerbosity 3 $ "File " ++ fn ++ " read.")
    case pM of
      Just m -> do
        -- imported module names
        let mods' = I.getImports m (fPrelude opts)
        let mods  = map makecall mods' \\ known
        let new_known = mods ++ known
        atVerbosity 3 $ "make: branching from " ++ frommod
        indir <- mapM (determineDependencies False opts new_known) mods
        (dir,fc) <- checkCompilation sp frommod
        -- all dependencies done; we can compile (if necessary)
        if   or indir || dir || top
         then case fc of
                Just (_,fromghs,fromghi,fromhs) ->
                          do atVerbosity 0 $ "make: compiling " ++ fromghs
                             compile opts frommod fromghs fromghi fromhs
                             return True
                Nothing ->
                          do atVerbosity 3 $ "make: no .ghs file for " ++ frommod
                             return False {- change here for Haskell compilation -}
         else do atVerbosity 3 $ "make: " ++ frommod ++ " unchanged"
                 return False
      _ -> return False
  where makecall (ModuleName_ModuleName _ (m:_))
                                 =  (m,m ++ ".ghs",m ++ ".ghi",m ++ ".hs")

\end{code}

The |compile| function directs the compilation process.
    
\begin{code}

compile                         :: Flags -> String -> String -> String -> String 
                                -> IO ()
compile opts bName
             modInName 
             modIName
             modOutName         =  do 
    -- verbosity currently depends on the number of times that
    -- the verbosity option is given on the command line
    let verbosity               =  fVerbose opts
    -- atVerbosity helps us to show certain output only if the
    -- verbosity level is high enough
    let atVerbosity n s         =  when (verbosity >= n) $ putStrLn s 
    atVerbosity 6 $ "DEBUG: modInName " ++ modInName
    atVerbosity 6 $ "DEBUG: modInName " ++ modOutName
    atVerbosity 6 $ "DEBUG: modInName " ++ modIName
    -- the search path is the collection of all paths given in
    -- library options
    sp                          <- expandPath $ reverse (fLibPath opts) ++ searchPath
    atVerbosity 4 $ "Using search path: " ++ concat (intersperse ":" sp)
    -- the maximum iteration number defaults to 50, but can be set
    -- as a command line option
    let max_it                  =  fCut opts
    -- check if we continue after errors
    let continue                =  fContinue opts
    -- check if we dump after parsing
    let dumpP                   =  if fDumpParse opts then
                                     case fDumpParseFile opts of
                                       Nothing -> Just (bName ++ ".parse.aterm")
                                       Just n  -> Just n
                                   else
                                     Nothing
    let dumpR                   =  if fDumpResolve opts then
                                     case fDumpResolveFile opts of
                                       Nothing -> Just (bName ++ ".resolve.aterm")
                                       Just n  -> Just n
                                   else
                                     Nothing
    let typecheck               =  fTypeCheck opts
    let optimize                =  fOptimize opts
    let poptimize               =  fPOptimize opts
    
    -- terminal width
    width                       <- getTerminalWidth

    -- *** Parse input file ***
    (parsedModule,fn)                  <- P.parseFile verbosity [] modInName
    case parsedModule of 
      Just m -> 
       do 
          -- Report succesful parsing
          atVerbosity 1 $ "File " ++ fn ++ " read."
          -- Dump ATerm
          if isJust dumpP then do
            writeFile (fromJust dumpP) (toATermString m)
            atVerbosity 1 $ "ATerm dumped."
           else do
          
          -- *** Process imports ***
          msys <- I.processImportList m verbosity (fPrelude opts) sp

          -- *** Resolve names ***
          (errs,stop,msys,eV,eT,eS,iE,hE,mods) <- return $ NR.resolve msys verbosity
          -- Print errors
          displayIO stdout (renderPretty 1.0 width errs)
          atVerbosity 2 $ "Name resolving complete."
          -- End program if there were errors
          if stop && not continue then exitFailure else do

	  -- *** Resolve operator precedences ***
	  (msys,stop,errs) <- return $ RO.resolveOperators msys
	  -- Print errors
	  displayIO stdout (renderPretty 1.0 width errs)
	  atVerbosity 2 $ "Operator precendence resolving complete."
          -- Dump ATerm
          if isJust dumpR then do
            writeFile (fromJust dumpR) (toATermString msys)
            atVerbosity 1 $ "ATerm dumped."
           else do
	  -- End program if there were errors
	  if stop && not continue then exitFailure else do

          -- *** Run Bastiaan's type checker ***
{- (removed for Coral because it currently doesn't work, ks, 06.01.2005)
	  when typecheck $ do
	    -- build initial environments
	    (i1,i2,i3,i4) <- return $ IT.prepare msys
	    -- wrapping
	    let (ModuleSystem_ModuleSystem _ tMod) = msys
	    (_,tDebug,tSErrors,_,toplevelKinds,_,tTErrors,tWarnings) 
	                    <- return $ SA.sem_Module tMod
		  	         modInName 
			         i1 i2 i3 i4
			         S.algW False
	    -- show debug output
	    if verbosity >= 2 then tDebug else return ()
	    -- show kind environment
            putStr . unlines $
                 map
                   (\(name,tp) -> show name ++ " :: " ++ show tp)
				  (MAL.toList toplevelKinds)
	    -- show type errors and warnings
	    putStrLn $ unlines $ SAM.sortAndShowErrors tSErrors
	    putStrLn $ unlines $ map show tTErrors
	    putStrLn $ unlines $ SAM.sortAndShowWarnings tWarnings
-}

          -- Analyse program
          -- * (errs,stop) <- return $ A.sem_Root root verbosity
          -- Print errors
          -- * putStr errs
          -- * atVerbosity 2 $ "Static analysis complete."
          -- End program if there were errors
          -- * if stop && not continue then exitFailure else do
          
          -- *** Load GHPrelude source 
{- (removed for Coral because it currently doesn't work, ks, 06.01.2005)
	  optimizerPrelude <-
	      if optimize 
	      then do atVerbosity 4 $ "Parsing GHPrelude for Optimizer"
		      decls1 <- ModuleLoader.loadModule verbosity sp 
				  Optimizer.filenameGHPrelude
		      decls2 <- ModuleLoader.loadModule verbosity sp 
				  Optimizer.filenameOptimizerPrelude
		      return (decls1 ++ decls2)
	      else return []
-}

          -- *** Call compiler ***
          atVerbosity 4 $ "Calling compiler ..."
          ((errs,stop,screen,file,ifile),_) 
              <- return $ GH.sem_Root 
                            (GH.Root_Root msys) 
                            (verbosity
                            ,fPrelude opts
                            ,max_it
                            ,fGHC opts
                            ,eV,eT,eS,iE,hE
                            ,mods
			    ,fOptimize opts
			    ,[] -- optimizerPrelude
                            )

          -- Print errors
          displayIO stdout (renderPretty 1.0 width errs)
          -- End program if necessary
          if stop && not continue then exitFailure else do
          -- Print screen output
          displayIO stdout (renderPretty 1.0 width screen)

          -- write interface file
          atVerbosity 4 $ "Writing interface: " ++ modIName
          ifH <- openFile modIName WriteMode
          displayIO ifH (renderPretty 0.9 80 ifile)
          hClose ifH
          atVerbosity 1 $ "Interface file " ++ modIName ++ " written."
          -- write output file
          atVerbosity 4 $ "Writing Haskell ..."
	  outH <- openFile modOutName WriteMode
          displayIO outH (renderPretty 0.9 80 file)
          hClose outH
          atVerbosity 1 $ "Haskell file " ++ modOutName ++ " written."

          -- ** Call optimizer ***
{- (removed for Coral because it currently doesn't work, ks, 06.01.2005)
          when poptimize $ do 
             atVerbosity 4 $ "Calling Optimizer ..."
             TestOptimizer.optimizeFile sp modOutName
-}

          {-
          -- pretty print input file to output
          if isJust dump then do
            writeFile (fromJust dump) (toATermString m)
            atVerbosity 1 $ "ATerm dumped."
           else do
          writeFile modOutName (($ "") . displayS . renderPretty 0.7 100 $ (PP.ppModule m))
          atVerbosity 1 $ "Pretty printed module " ++ modOutName ++ " written."
	  -}

      _ -> {- we had parse errors -}
           exitFailure
\end{code}

%---------------------------------------------------------------------------
\subsection{The main function}
%---------------------------------------------------------------------------

The |main| function retrieves the input file name from the
command line or asks for it, then it calls the compiler.

\begin{code}

main                            :: IO ()
main                            =  do
                                     -- we introduce ourselves 
                                     -- putStrLn signon
                                     -- we get the command line
                                     args <- getArgs 
                                     -- we get the program name
                                     progName <- getProgName
                                     -- give the work to another function
                                     mainWithNameAndArgs progName args

mainWithNameAndArgs             :: String -> [String] -> IO ()  
mainWithNameAndArgs progName args =
  do
    -- analyse command line options
    let opts = selectOptions progName args
    case opts of
      -- we had errors
      Right s      -> do putStrLn s
      -- no file specified
      Left  (o,[]) -> callMode 
                        ( o { fVerbose     = max (fVerbose o) 1  
                            , fInteractive = True
                            }
                        )
      Left  (o,n ) -> callMode
                        ( o { fInputs      = n } )

\end{code}

In interactive mode, ask for filename.

\begin{code}

callMode                        :: Flags -> IO ()
callMode opts =
  case fCallMode opts of
    GetVersion -> displayVersionInfo
    PrintSearchPath
               -> do let ps  =   searchPath
                     putStrLn "before expansion:"
                     print ps
                     eps     <-  expandPath $ 
                                   reverse (fLibPath opts) ++ searchPath
                     putStrLn "after expansion:"
                     print eps
                     putStrLn ""
    GetHelp _  -> displayUsageInfo
    _ | fInteractive opts
               -> do
                    displayVersionInfo
                    putStr $ "Enter filename: "
                    hFlush stdout
                    name <- getLineWithBackspaces
                    findAndCompile opts name
      | otherwise
               -> do
                    sequence_ (map (findAndCompile opts) (fInputs opts))

\end{code}                    

Functions displaying version and program usage info.

\begin{code}

displayVersionInfo              :: IO ()
displayVersionInfo              =  putStrLn signon

displayUsageInfo                :: IO ()
displayUsageInfo                =  putStr $ usageInfo uheader (concatMap removeOption options)

-- we're keeping --no-typecheck for compatibility, but don't display it
-- in Coral; we also use this function to hide the aterm-options
removeOption                    :: OptDescr a -> [OptDescr a]
removeOption (Option _ (l:_) _ _)
  | l `elem` hiddenOptions      =  []
removeOption o                  =  [o]

signon,vinfo                    :: String        
signon                          =  "Generic Haskell compiler/UHA, version "
                                ++ vinfo
vinfo                           =  version ++ ", " ++ operatingSystem ++ " build"

{- 
  color release possibilities:
    Amber
    Blue, Brown, Beryl
    Coral, Crimson, Cyan, Citrine
    Diamond (?)
    Emerald
-}

uheader                         :: String
uheader                         =  "Usage: gh [options...] files..."

\end{code}

%---------------------------------------------------------------------------
\subsection{Option handling}
%---------------------------------------------------------------------------

There is nothing special about the options. Only thing to note: You can
specify |Verbose| as an option multiple times. The output is supposed to
be more verbose then. A general guideline is:

\begin{center}
\begin{tabular}{ll} \hline
verbosity level & output \\ \hline
0 & as quiet as possible, only errors are shown \\
1 & display warnings and errors, minor diagnostic output \\
2 & display some intermediate results \\
3+ & debug output \\ \hline
\end{tabular}
\end{center}

\begin{code}

selectOptions                   :: String 
                                -> [String] 
                                -> Either (Flags,[String]) -- options
                                          String            -- error
selectOptions progName args     = 
  case getOpt Permute options args of
    (o,n,[]  ) -> let flags = foldr ($) stdFlags o
                  in 
                    case fCallMode flags of
                      Error s -> quitError s
                      _       -> Left  $ (flags,n)
    (_,_,errs) -> quitError (concat errs)
 where
   quitError s = Right $ s ++ usageInfo uheader options

data Flags                      =  Flags
                                     { fCallMode    :: CallMode
                                     , fVerbose     :: Int
                                     , fInteractive :: Bool
                                     , fTypeCheck   :: Bool
                                     , fOutput      :: Maybe String
                                     , fCut         :: Int
                                     , fContinue    :: Bool
                                     , fLibPath     :: [String]
                                     , fDumpParse   :: Bool
                                     , fDumpParseFile :: Maybe String
                                     , fDumpResolve :: Bool
                                     , fDumpResolveFile :: Maybe String
                                     , fGHC         :: Bool
                                     , fPrelude     :: Bool
                                     , fInputs      :: [String]
				     , fOptimize    :: Bool
                                     , fPOptimize   :: Bool
                                     }

data CallMode                   =  GetHelp Bool | GetVersion | Make | Compile | PrintSearchPath
                                |  Error String

stdFlags                        =  Flags
                                     { fCallMode    =  Compile
                                     , fVerbose     =  0
                                     , fInteractive =  False
                                     , fTypeCheck   =  False
                                     , fOutput      =  Nothing
                                     , fCut         =  50
                                     , fContinue    =  False
                                     , fLibPath     =  []
                                     , fDumpParse   =  False
                                     , fDumpParseFile = Nothing
                                     , fDumpResolve =  False
                                     , fDumpResolveFile = Nothing
                                     , fGHC         =  True
                                     , fPrelude     =  False
                                     , fInputs      =  []
                                     , fOptimize    =  False
                                     , fPOptimize    =  False
                                     }

updateCallMode                  :: Flags -> CallMode -> Flags
updateCallMode f m              =  case fCallMode f of
                                     Compile -> f { fCallMode = m }
                                     _       -> f { fCallMode = Error 
                                                      "only one of -v, -m, -h allowed\n"
                                                  }

incVerbosity                    :: Flags -> Flags
incVerbosity f                  =  f { fVerbose = fVerbose f + 1 }


options                         :: [OptDescr (Flags -> Flags)]
options                         =
  [ Option ['v']     ["verbose"]       (NoArg incVerbosity)
                                       "(number of v's controls the verbosity)"
  , Option ['m']     ["make"]          (NoArg (\f -> updateCallMode f Make))
                                       "compile dependent files first"
  , Option ['V'] ["version","release"] (NoArg (\f -> updateCallMode f GetVersion))
                                       "show version info"
  , Option ['h','?'] ["help"]          (NoArg (\f -> updateCallMode f (GetHelp True)))
                                       "show help"
  , Option ['C']     ["continue"]      (NoArg (\f -> f { fContinue = True }))
                                       "continue after errors"
  , Option ['L']     ["library"]       (ReqArg (\p f -> f { fLibPath = p : fLibPath f })
                                               "DIR"
                                       )
                                       "add DIR to search path"
  , Option []        ["ghc"]           (NoArg (\f -> f { fGHC = True }))
                                       "use GHC extensions in output (default)"
  , Option []        ["no-ghc"]        (NoArg (\f -> f { fGHC = False }))
                                       "use no GHC extensions in output"
  , Option []        ["odir"]          (ReqArg (\n f -> f { fOutput = Just n }) "DIR")
                                       "place output files in DIR"
  , Option []        ["dump-aterm-after-parsing"]  
                                       (OptArg (\n f -> f { fDumpParse = True
                                                          , fDumpParseFile = n
                                                          }
                                               )
                                               "FILE"
                                       )
                                       "write ATerm to FILE after parsing"
  , Option []        ["dump-aterm-after-resolving"]  
                                       (OptArg (\n f -> f { fDumpResolve = True
                                                          , fDumpResolveFile = n
                                                          }
                                               )
                                               "FILE"
                                       )
                                       "write ATerm to FILE after name resolving"
  , Option []        ["print-searchpath"]
                                       (NoArg (\f -> updateCallMode f PrintSearchPath))
                                       "print search path"
--  , Option []        ["typecheck"]     (NoArg (\f -> f { fTypeCheck = True }))
--                                       "enable typechecker (EXPERIMENTAL!)"
  , Option []        ["no-typecheck"]  (NoArg (\f -> f { fTypeCheck = False }))
                                       "skip type checking (default)"
  , Option []        ["prelude"]       (NoArg (\f -> f { fPrelude = True }))
                                       "compile the prelude"
  , Option ['c']     ["cut"]           (ReqArg (\n f -> f { fCut = read n }) "N")
                                       "cut after N iterations (default: 50)"
--  , Option []        ["optimize"]      (NoArg (\f -> f { fOptimize = True }))
--                                       "generate optimized code (EXPERIMENTAL!)"
--  , Option []        ["poptimize"]     (NoArg (\f -> f { fPOptimize = True }))
--                                       "post-optimize code (EXPERIMENTAL!)"
 -- , Option ['o']     ["output"]        (OptArg out "FILE")  
 --   "place output in FILE"
  ]

hiddenOptions = ["typecheck","no-typecheck","optimize","poptimize",
                 "dump-aterm-after-parsing","dump-aterm-after-resolving",
                 "cut","prelude"]

\end{code}
