%% gh -- a compiler for Generic Haskell.
%% Copyright (c) 2001 -- 2005  The Generic Haskell Team. Utrecht University
%% 
%% This library is free software; you can redistribute it and/or modify
%% it under the terms of the GNU Lesser General Public License as
%% published by the Free Software Foundation; either version 2.1 of the
%% License, or (at your option) any later version.
%% 
%% This library is distributed in the hope that it will be useful, but
%% WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.
%% 
%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
%%
%% $Id: Name.lhs 194 2005-01-12 09:16:08Z andres $
%%
%% author: andres@cs.uu.nl

> module Name where
>
> import Char
> import Utils
> import AbstractSyntax
> import UHA_Range
> import Monad (liftM)

< import FMap

This module contains common functions used for creating the names
of the generated functions and types.

%***************************************************************************
\section{Names}
%***************************************************************************

> extractName                   :: Name -> String
> extractName (Name_Identifier _ _ n) = n
> extractName (Name_Operator _ _ n)   = n
> extractName (Name_Special _ _ n)    = n
> extractName (Name_Entity _ n _) = extractName n

> getModuleName                 :: ModuleName -> [String]
> getModuleName (ModuleName_ModuleName _ n) = n

> isSpecial                     :: Name -> Bool
> isSpecial (Name_Special _ _ _) = True
> isSpecial _                    = False

> isOperator                     :: Name -> Bool
> isOperator (Name_Operator _ _ _) = True
> isOperator _                     = False

> makeExplicitName                       :: Bool -> Name -> Name
> makeExplicitName False n               =  n
> makeExplicitName True (Name_Entity r e _) = Name_Entity r e MaybeName_Nothing
> makeExplicitName _ n                   =  n

> makePlainName                          :: Name -> Name
> makePlainName (Name_Identifier _ _ n)  =  Name_Identifier unknownRange [] n
> makePlainName (Name_Operator _ _ n)    =  Name_Operator   unknownRange [] n
> makePlainName (Name_Special _ _ n)     =  Name_Special    unknownRange [] n
> makePlainName (Name_Entity _ n _)      =  makePlainName n

> isQualifiedName               :: Name -> Bool
> isQualifiedName (Name_Identifier _ [] _) =  False
> isQualifiedName (Name_Operator   _ [] _) =  False
> isQualifiedName (Name_Special    _ [] _) =  False
> isQualifiedName (Name_Entity _ _ (MaybeName_Just n))
>                                          =  isQualifiedName n
> isQualifiedName _                        =  True

> knownEntity                   :: Name -> Name
> knownEntity n                 =  Name_Entity (getRange n) n MaybeName_Nothing {- (MaybeName_Just n) -}

> knownSpecialEntity            :: Name -> Name
> knownSpecialEntity n          =  Name_Entity (getRange n) n (MaybeName_Just n)

The following functions allow to distinguish variables from constructors.

> isCon, isVar                  :: Name -> Bool
> isCon                         =  isConid . extractName
> isVar                         =  isVarid . extractName

Extract the name of a function binding from its left hand side.

> nameFromLhs :: LeftHandSide -> Name
> nameFromLhs (LeftHandSide_Parenthesized _ l _) = nameFromLhs l
> nameFromLhs (LeftHandSide_Function      _ n _) = n
> nameFromLhs (LeftHandSide_Infix       _ _ o _) = o

Module prefix.

> moduleName                    :: String -> String
> moduleName                    =  (modulePrefix++)

> modulePrefix                  :: String
> modulePrefix                  =  "GHM_"

Name of the structural type.

> strName                       :: Name -> Name
> strName                       =  localDerivedName ("GHt_"++)

Name of the structural specialisation for a type-indexed type.

> strTITName                    :: Name -> Name
> strTITName                    =  strName

Name of embedding projection pair.

> epName                        :: Name -> Name
> epName                        =  localDerivedName ("ghEP_"++)

> epTypeName,epfromName,eptoName,idEPName,liftEPName
>                               :: Name
> epTypeName                    =  inPrelude "EP"
> epfromName                    =  inPrelude "from"
> eptoName                      =  inPrelude "to"
> idEPName                      =  inPrelude "idEP"
> liftEPName                    =  inPrelude "liftEP"

Name of bimap.

> bimapName                     :: Name
> bimapName                     =  inPrelude "bimap"

We distinguish component and specialisation names. Whereas
components are always located in the same module where the generic function
is declared, specialisations are always local to the current module.

> componentName                 :: Name -> Name -> Name
> componentName fun ctor        =  let mangled_fun  = extractName fun
>                                      mangled_ctor = mangleName ctor
>                                      basename     = "ghc_" ++ mangleBigPair mangled_fun mangled_ctor
>                                  in  localDerivedName (const basename) fun

> specName                      :: Name -> Name -> Name
> specName fun ctor             =  let mangled_fun  = mangleName fun
>                                      mangled_ctor = mangleName ctor
>                                      basename     = "ghs_" ++ mangleBigPair mangled_fun mangled_ctor
>                                  in  Name_Identifier unknownRange [] basename

> kitName                       :: Name -> Name
> kitName                       =  localDerivedName ("GHk_"++)

The same for type-indexed data types.

> componentTITName              :: Name -> Name -> Name
> componentTITName fun ctor     =  let mangled_fun  = extractName fun
>                                      mangled_ctor = mangleName ctor
>                                      basename     = "GHc_" ++ mangleBigPair mangled_fun mangled_ctor
>                                  in  localDerivedName (const basename) fun

> specTITName                   :: Name -> Name -> Name
> specTITName fun ctor          =  let mangled_fun  = mangleName fun
>                                      mangled_ctor = mangleName ctor
>                                      basename     = "GHs_" ++ mangleBigPair mangled_fun mangled_ctor
>                                  in  Name_Identifier unknownRange [] basename

Similar for generic abstractions.

> specGAName                    :: Name -> Type -> Name
> specGAName fun tp             =  let mangled_fun  = mangleName fun
>                                      mangled_tp   = mangleType tp
>                                      basename     = "ghs_" ++ mangleBigPair mangled_fun mangled_tp
>                                  in  Name_Identifier unknownRange [] basename

Name of the structure type constructors (and their value constructors).

> unitName,unitunitName,conName,conconName,
>  labelName,labellabelName,sumName,sumleftName,
>  sumrightName,prodName,prodprodName,funName,listName,pairName,
>  condescrName,condescrcondescrName,labeldescrName,
>  labeldescrlabeldescrName
>                               :: Name
> unitName                      =  inPrelude "Unit"
> unitunitName                  =  inPrelude "Unit"
> conName                       =  inPrelude "Con"
> conconName                    =  inPrelude "Con"
> labelName                     =  inPrelude "Label"
> labellabelName                =  inPrelude "Label"
> sumName                       =  inPrelude "Sum"
> sumleftName                   =  inPrelude "Inl"
> sumrightName                  =  inPrelude "Inr"
> prodName                      =  inPrelude "Prod"
> prodprodName                  =  inPreludeSymbol ":*:"
> funName                       =  inPrelude "Fun"

Generic identity function, generic identity type.

> gidName, gIDName              :: Name
> gidName                       =  inPrelude "gid"
> gIDName                       =  inPrelude "GID"

Builtin types.

> listName                      =  knownSpecialEntity $ Name_Special unknownRange hsPrelude "[]"
> pairName                      =  tupleName 2
> tupleName                     :: Int -> Name
> tupleName n                   =  knownSpecialEntity $ Name_Special unknownRange hsPrelude
>                                                  ("(" ++ replicate (n-1) ',' ++ ")" )

Constructor descriptions.

> condescrName                  =  inPrelude "ConDescr"
> condescrcondescrName          =  inPrelude "ConDescr"

Label descriptions.

> labeldescrName                =  inPrelude "LabelDescr"
> labeldescrlabeldescrName      =  inPrelude "LabelDescr"

Constructor case prefix

> concaseName                   :: Name -> Name
> concaseName                   =  localDerivedName ("GHC_"++)

Constructor descriptors of specific constructors.

> cdName                        :: Name -> Name
> cdName                        =  localDerivedName ("ghC_"++)

Label descriptors of specific labels.

> ldName                        :: Name -> Name
> ldName                        =  localDerivedName ("ghL_"++)

Kind case name.

> kindcaseName                  :: Name
> kindcaseName                  =  Name_Identifier unknownRange [] "GHK_S"

For "short notation" generic abstraction specializations, we need a canonical
way to convert a kind into a name.

> kindToName                           :: Kind -> Name
> kindToName k                         =  Name_Identifier unknownRange [] ("GHK_" ++ t k)
>   where  t (Kind_Star _)             =  "S"
>          t (Kind_Arrow _ l r)        =  "A" ++ t l ++ t r
>          t (Kind_Parenthesized _ k)  =  t k
>          t _                         =  "" -- or error

Abstract type name. Used for generic abstractions.
Since it does not occur in any generated piece of code, any
unique name will do.

> absName                       :: Name
> absName                       =  Name_Identifier unknownRange [] "GHK_ABS"

Variables in specialised types (hack).

> varTITName                    :: Name -> Name -> Name
> varTITName                    =  specName

Name of constructor descriptor's fixity constructors.

> nonfixName,infixName,infixlName,infixrName
>                               :: Name
> nonfixName                    =  inPrelude "Nonfix"
> infixName                     =  inPrelude "Infix"
> infixlName                    =  inPrelude "Infixl"
> infixrName                    =  inPrelude "Infixr"

Checks whether an (unqualified) name is a valid unit or tuple constructor.

> isUnitName,isTupleName        :: Name -> Bool
> isUnitName (Name_Special _ p "()")  
>   | null p || p == hsPrelude        = True
> isUnitName (Name_Entity _ n _)      = isUnitName n
> isUnitName _                        = False
> isTupleName (Name_Special _ p n)
>   | (null p || p == hsPrelude)
>        && isTupleString n           = True
> isTupleName (Name_Entity _ n _)     = isTupleName n
> isTupleName _                       = False

(Real) Haskell names/basic values.

> hsPrelude                     :: [String] -- Modid
> hsPrelude                     =  ["Prelude"]
> hsPreludeModule               :: ModuleName
> hsPreludeModule               =  ModuleName_ModuleName unknownRange hsPrelude
> hsUnitName                    :: Name
> hsUnitName                    =  knownSpecialEntity $ Name_Special unknownRange hsPrelude "()"
> hsFunName                     :: Name
> hsFunName                     =  knownSpecialEntity $ Name_Special unknownRange hsPrelude "->"
> hsUnitConName                 :: Name
> hsUnitConName                 =  knownSpecialEntity $ Name_Special unknownRange hsPrelude "()"
> hsListName                    :: Name
> hsListName                    =  knownSpecialEntity $ Name_Special unknownRange hsPrelude "[]"
> hsListConName                 :: Name
> hsListConName                 =  knownSpecialEntity $ Name_Special unknownRange hsPrelude ":"
> hsTupleName                   :: Int -> Name
> hsTupleName n                 =  knownSpecialEntity
>                               $  Name_Special unknownRange hsPrelude
>                                               ("(" ++ replicate (n-1) ',' ++ ")")
> hsTupleConName                :: Int -> Name
> hsTupleConName n              =  knownSpecialEntity
>                               $  Name_Special unknownRange hsPrelude
>                                               ("(" ++ replicate (n-1) ',' ++ ")")
> hsMinusName                   :: Range -> Strings -> Name
> hsMinusName r m               =  Name_Operator r m "-"
> hsPlingName                   :: Range -> Strings -> Name
> hsPlingName r m               =  Name_Operator r m "!"
> hsDotName                     :: Range -> Strings -> Name
> hsDotName r m                 =  Name_Operator r m "."
> hsAsName                      :: Range -> Strings -> Name
> hsAsName r m                  =  Name_Identifier r m "as"
> hsQualifiedName               :: Range -> Strings -> Name
> hsQualifiedName r m           =  Name_Identifier r m "qualified"
> hsHidingName                  :: Range -> Strings -> Name
> hsHidingName r m              =  Name_Identifier r m "hiding"
> hsExtendsName                 :: Range -> Strings -> Name
> hsExtendsName r m             =  Name_Identifier r m "extends"
> hsBoolName                    :: Bool -> Name
> hsBoolName False              =  knownEntity $ Name_Identifier unknownRange hsPrelude "False"
> hsBoolName True               =  knownEntity $ Name_Identifier unknownRange hsPrelude "True"

Variable list (in cases where the name space is not dangerous).

> varNames                      :: [Name]
> varNames                      =  map (Name_Identifier unknownRange []) varNames'
> varNames'                     =  map (:[]) alphabet
>                               ++ [ y:x | x <- varNames', y <- alphabet ]
>   where alphabet              =  ['a'..'z']

Prefix a variable with a rare prefix to prevent name clashes.

> prefixName                    :: Name -> Name
> prefixName                    =  foreignDerivedName ("gh_"++)

Impossible name.

> impossibleName                :: Name
> impossibleName                =  Name_Entity unknownRange 
>                                              (Name_Special unknownRange [] "*<impossible>")
>                                              MaybeName_Nothing

Prefixed variable list.

> prefixedVarNames              :: [Name]
> prefixedVarNames              =  map prefixName varNames

Name of the GH prelude.

> preludeName                   :: Strings -- Modid
> preludeName                   =  ["GH","Prelude"]

> preludeModule                 :: ModuleName
> preludeModule                 =  ModuleName_ModuleName unknownRange preludeName

Functions of the GH prelude.

> inPrelude                     :: String -> Name
> inPrelude n                   =  knownEntity $ (Name_Identifier unknownRange preludeName n)

> inPreludeSymbol               :: String -> Name
> inPreludeSymbol s             =  knownEntity $ (Name_Operator unknownRange preludeName s)

> localName                     :: String -> Name
> localName                     =  Name_Identifier unknownRange []

E-mail address for bug reports.

> email                         =  "bugs@generic-haskell.org"

Derived names and mangling functions. Should be documented ;-)

> foreignDerivedName :: (String -> String) -> Name -> Name
> foreignDerivedName f n = Name_Identifier 
>                            unknownRange 
>                            (foreignDerivedModuleName (getModule n))
>                            (f . mangleName $ n)

> localDerivedName :: (String -> String) -> Name -> Name
> localDerivedName f n   = Name_Identifier
>                            unknownRange
>                            (foreignDerivedModuleName (getModule n))
>                            (f . mangleName . makePlainName $ n)

> getModule                         :: Name -> Strings
> getModule (Name_Identifier r m n) =  m
> getModule (Name_Operator   r m n) =  m
> getModule (Name_Special    r m n) =  m
> getModule (Name_Entity     r n s) =  getModule n

> foreignDerivedModuleName          :: Strings -> Strings
> foreignDerivedModuleName m | m == hsPrelude      = preludeName
>                            | otherwise           = m

> mangleName                         :: Name -> String
> mangleName (Name_Identifier _ m n) =  mangleModule m n
> mangleName (Name_Operator   _ m n) =  mangleModule m (manglePrefix sprefix n)
>   where sprefix = if isConid n then "GHS_" else "ghS_"
> mangleName (Name_Special    _ m n) =  {- mangleModule m -}
>                                         (manglePrefix xprefix n)
>   where xprefix = if isConid n then "GHX_" else "ghX_"
> mangleName (Name_Entity     _ n _) =  mangleName n

> mangleModule :: Strings -> String -> String
> mangleModule m n = mangleList (m ++ [n])

> manglePrefix                       :: String -> String -> String
> manglePrefix prefix s              =  prefix ++ concatMap mangleSymbol s

> mangleBigPair            :: String -> String -> String
> mangleBigPair p1 p2      =  escape p1 ++ "__" ++ p2
>   where escape ('_':'_':r) = "q_"  ++ escape r
>         escape ('q':'_':r) = "qq_" ++ escape r
>         escape ('q':'q':r) = "qq"  ++ escape ('q':r)
>         escape (s:r)       = s : escape r
>         escape []          = []

> unmangleBigPair          :: String -> (String,String)
> unmangleBigPair          =  stripFirst ""
>   where stripFirst acc ('_':'_':r) = (reverse acc,r)
>         stripFirst acc ('q':'q':r) = stripFirst ('q':acc) r
>         stripFirst acc ('q':'_':r) = stripFirst ("__" ++ acc) r
>         stripFirst acc (n      :r) = stripFirst (n:acc) r
>         stripFirst acc []          = (reverse acc,"") -- error tolerance

> manglePair               :: String -> String -> String
> manglePair p1 p2         =  escape p1 ++ "_" ++ p2
>   where escape           =  concatMap escapeChar
>         escapeChar '_'   =  "q_"
>         escapeChar 'q'   =  "qq"
>         escapeChar  x    =  [x]

> unmanglePair             :: String -> (String,String)
> unmanglePair             =  stripFirst ""
>   where stripFirst acc ('_' : r) = (reverse acc,r)
>         stripFirst acc []        = (reverse acc,"")  -- error tolerance
>         stripFirst acc ('q':n:r) = stripFirst (n:acc) r
>         stripFirst acc (n   : r) = stripFirst (n:acc) r

> mangleList               :: [String] -> String
> mangleList               =  foldr1 manglePair

> unmangleList             :: Int -> String -> [String]
> unmangleList 1 p         =  [p]
> unmangleList n p         =  let (p1,p2) = unmanglePair p
>                             in  p1 : unmangleList (n-1) p2

> mangleSymbol             :: Char -> String
> mangleSymbol '&'         =  "ad"
> mangleSymbol '$'         =  "ap"
> mangleSymbol '@'         =  "at"
> mangleSymbol ']'         =  "bc"
> mangleSymbol '\\'        =  "bk"
> mangleSymbol '['         =  "bo"
> mangleSymbol '`'         =  "bq"
> mangleSymbol '}'         =  "cc"
> mangleSymbol ':'         =  "cl"
> mangleSymbol ','         =  "cm"
> mangleSymbol '{'         =  "co"
> mangleSymbol '.'         =  "dt"
> mangleSymbol '/'         =  "dv"
> mangleSymbol '='         =  "eq"
> mangleSymbol '!'         =  "ex"
> mangleSymbol '>'         =  "gt"
> mangleSymbol '#'         =  "hs"
> mangleSymbol '<'         =  "lt"
> mangleSymbol '%'         =  "md"
> mangleSymbol '-'         =  "mn"
> mangleSymbol '?'         =  "op"
> mangleSymbol '|'         =  "or"
> mangleSymbol ')'         =  "pc"
> mangleSymbol '+'         =  "pl"
> mangleSymbol '('         =  "po"
> mangleSymbol '^'         =  "pw"
> mangleSymbol '"'         =  "qu"
> mangleSymbol ';'         =  "sc"
> mangleSymbol '~'         =  "tl"
> mangleSymbol '*'         =  "tm"
> mangleSymbol c           =  "N" ++ [c]

> unmangle                 :: String -> Range -> Strings -> Name
> unmangle ('G':'H':'S':'_':rest) r m = Name_Operator r m $ unmangle' rest
> unmangle ('g':'h':'S':'_':rest) r m = Name_Operator r m $ unmangle' rest
> unmangle ('G':'H':'X':'_':rest) r m = Name_Special  r m $ unmangle' rest
> unmangle ('g':'h':'X':'_':rest) r m = Name_Special  r m $ unmangle' rest
> unmangle x r m           =  Name_Identifier r m x
>
> unmangle' []             =  []
> unmangle' [x]            =  [x]
> unmangle' (x:y:rest)     = unmangleSymbol [x,y] : unmangle' rest
>
> unmangleSymbol           :: String -> Char
> unmangleSymbol "ad"      =  '&'
> unmangleSymbol "ap"      =  '$'
> unmangleSymbol "at"      =  '@'
> unmangleSymbol "bc"      =  ']'
> unmangleSymbol "bk"      =  '\\'
> unmangleSymbol "bo"      =  '['
> unmangleSymbol "bq"      =  '`'
> unmangleSymbol "cc"      =  '}'
> unmangleSymbol "cl"      =  ':'
> unmangleSymbol "cm"      =  ','
> unmangleSymbol "co"      =  '{'
> unmangleSymbol "dt"      =  '.'
> unmangleSymbol "dv"      =  '/'
> unmangleSymbol "eq"      =  '='
> unmangleSymbol "ex"      =  '!'
> unmangleSymbol "gt"      =  '>'
> unmangleSymbol "hs"      =  '#'
> unmangleSymbol "lt"      =  '<'
> unmangleSymbol "md"      =  '%'
> unmangleSymbol "mn"      =  '-'
> unmangleSymbol "op"      =  '?'
> unmangleSymbol "or"      =  '|'
> unmangleSymbol "pc"      =  ')'
> unmangleSymbol "pl"      =  '+'
> unmangleSymbol "po"      =  '('
> unmangleSymbol "pw"      =  '^'
> unmangleSymbol "qu"      =  '"'
> unmangleSymbol "sc"      =  ';'
> unmangleSymbol "tl"      =  '~'
> unmangleSymbol "tm"      =  '*'

> mangleType               :: Type -> String
> mangleType t             =  "GHT_" ++ mangleType' t
>   where mangleType'                     :: Type -> String
>         mangleType' (Type_Application _ _ t ts) 
>           =  mangleList (["a",mangleType' t] ++ map mangleType' ts)
>         mangleType' (Type_Variable _ n)
>           =  mangleList ["v",mangleName n]
>         mangleType' (Type_Constructor _ n)
>           =  mangleList ["c",mangleName n]
>         mangleType' (Type_Qualified _ _ _)
>           =  "<error:qualified>"
>         mangleType' (Type_Forall _ _ _)
>           =  "<error:forall>"
>         mangleType' (Type_Exists _ _ _)
>           =  "<error:exists>"
>         mangleType' (Type_Parenthesized _ t)
>           =  mangleType' t
>         mangleType' (Type_TypeApplication _ _ _)
>           =  "<error:typeapp>"
>         mangleType' (Type_KindApplication _ _ _)
>           =  "<error:kindapp>"
>         mangleType' (Type_ConstructorCase _ n)
>           =  mangleList ["C",mangleName n]

