-- gh -- a compiler for Generic Haskell.
-- Copyright (c) 2001 -- 2005  The Generic Haskell Team. Utrecht University
-- 
-- This library is free software; you can redistribute it and/or modify
-- it under the terms of the GNU Lesser General Public License as
-- published by the Free Software Foundation; either version 2.1 of the
-- License, or (at your option) any later version.
-- 
-- This library is distributed in the hope that it will be useful, but
-- WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-- Lesser General Public License for more details.
-- 
-- You should have received a copy of the GNU Lesser General Public
-- License along with this library; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
--
-- $Id: RequestUtils.hs 194 2005-01-12 09:16:08Z andres $
--
-- author: Andres Loeh (andres@cs.uu.nl)

module RequestUtils where

import AbstractSyntax
import Types
import UHA_Range
import BuildSyntax
import Name

import FMap
import Pretty
import ErrorUtils
import ErrorMessages
import qualified Pretty as PP

import Set
import qualified List
import List hiding (find)
import Maybe
import PPrint

newtype Request_TypeImply        =  Request_TypeImply  (Type,[Type])
newtype Request_BimapImply       =  Request_BimapImply (Name,Type)
newtype Request_FunImply         =  Request_FunImply   (Name
                                                       ,[(Name,RawRequest)]
                                                       )
newtype RawRequest               =  RawRequest         (Name,Type)
newtype DepConstraint            =  DepConstraint      (Range,Name,Name)

instance Eq DepConstraint where
  DepConstraint (_,n1,v1) == DepConstraint (_,n2,v2) = n1 == n2 && v1 == v2

newtype ComponentRequest         =  ComponentRequest
                                      (Name   -- of generic function
                                      ,Name   -- of type
                                      ,KindIndexedType -- of function
                                      )
newtype ComponentTITRequest      =  ComponentTITRequest
                                      (Name   -- of type-indexed type
                                      ,Name   -- of type
                                      )
newtype KindCaseRequest          =  KindCaseRequest
                                      (Name   -- of generic function
                                      ,Name   -- of type
                                      ,KindIndexedType -- of function
                                      )
newtype GenericAbstractionRequest=  GenericAbstractionRequest
                                      (Name   -- of generic function
                                      ,Type
                                      )
newtype NormalRequest            =  NormalRequest
                                      (Name   -- of generic function
                                      ,Name   -- of type
                                      ,LambdaType -- structure type
                                      ,KindIndexedType -- of function
                                      )
newtype NormalTITRequest         =  NormalTITRequest
                                      (Name   -- of type-indexed type
                                      ,Name   -- of type
                                      ,LambdaType -- structure type
                                      )

-- ks, 10.12.2004: equality for requests -- the |matchType| function
-- obviously fails to match any two variables, whereas |RawRequests|
-- for variables are either unresolved dependencies which lead to failure
-- or generic abstraction requests, which are all equal; I don't want
-- to create any new errors, so I introduce a new function |bothVar|.

instance Eq RawRequest where
  RawRequest (n1,t1) == RawRequest (n2,t2) 
                                 =  n1 == n2 && 
                                    (t1 `matchType` t2 || t1 `bothVar` t2)

bothVar :: Type -> Type -> Bool
bothVar (Type_Variable _ _)  (Type_Variable _ _)  =  True
bothVar _                    _                    =  False

type Request                     =  (Name,Declaration,KindIndexedType)
                    {- (generic function's
                       ,declaration of structure type to specialise to
                       ,generic function's kind-indexed type
                       )
                    -}
type Requests                    =  [Request]


request_iterate                  :: Int                -- how many iterations?
                                 -> [RawRequest]       -- new requests
                                 -> FMap Name Kind     -- generic abstractions
                                 -> [Request_TypeImply]
                                 -> [Request_BimapImply]
                                 -> [Request_FunImply]
                                 -> FMap Name [TypeArgument] -- consigs
                                 -> FMap Name Kind     -- kinds
                                 -> (Int               -- iterations left
                                    ,[RawRequest] -- final generic abs rqsts
                                    ,[RawRequest] -- final normal/MPC requests
                                    )

request_iterate n newRequests genericAbstractions
                typeImps bimapImps funImps consigs kinds =
    let rqs@(gaRequests,mpcRequests) = clean_requests newRequests
    in  request_fix n rqs rqs
 where
  request_fix                    :: Int           -- iteration count
                                 -> ([RawRequest],[RawRequest]) 
                                                  -- accumulator (GA,MPC)
                                 -> ([RawRequest],[RawRequest])
                                                  -- new (GA,MPC) 
                                 -> (Int          -- new iteration count
                                    ,[RawRequest] -- new GA requests
                                    ,[RawRequest] -- new MPC requests
                                    )
  request_fix n (agas,ampcs) ([],[]) = (n,agas,ampcs)
    -- If there is nothing new, we are done

  request_fix 0 (agas,ampcs) _   =  (-1,agas,ampcs)
    -- We have run out of iterations, we set the iteration count to
    -- $-1$ to indicate that.

  request_fix n (agas,ampcs) (gas,mpcs) =
    let new_from_mpcs            =  bimap_implications mpcs 
                                      ++ type_implications mpcs 
                                      ++ fun_implications mpcs
        new_from_gas             =  fun_implications gas
        (cleaned_gas,cleaned_mpcs)  
                                 =  clean_requests 
                                      (new_from_mpcs ++ new_from_gas) 
        (realnew_gas,realnew_mpcs)
                                 =  (cleaned_gas   \\  agas
                                    ,cleaned_mpcs  \\  ampcs
                                    )
    in  request_fix 
          (n-1)                      -- decrease iteration counter
          (realnew_gas  ++ agas      -- add the new ones ...
          ,realnew_mpcs ++ ampcs     -- to the alreadt accumulated
          ) 
          (realnew_gas,realnew_mpcs) -- these are the new ones
      
  clean_requests                 :: [RawRequest] 
                                 -> ([RawRequest] -- generic abs requests
                                    ,[RawRequest] -- normal/MPC requests
                                    )
  clean_requests requests        =
    let (gas,mpcs)               =  List.partition 
                                      (\(RawRequest (name,tp)) ->
                                        isJust (find genericAbstractions name))
                                      requests
        nubbed_gas               =  List.nub gas
        splitted_mpcs            =  List.nub 
                                      [ RawRequest (name,x)
                                      | RawRequest (name,tp) <- mpcs
                                      , x <- splitType tp
                                      ]
    in  (nubbed_gas,splitted_mpcs)

  type_implications              :: [RawRequest] -> [RawRequest]
  type_implications requests     =
    [ RawRequest (name,x)
    | RawRequest (name,tp)       <- requests
    , let consigs_for_name       =  applyFMM consigs (const []) name
    , Maybe.isNothing 
        (List.find 
           (\other -> matchType tp (typeArgumentToType other))
           consigs_for_name
        )
            -- the request should not be a component request
    , not 
        ( ( case applyFMM 
                   kinds 
                   (const $ Kind_Star unknownRange) 
                   (typeName tp) 
            of
              Kind_Star _          -> True
              _                    -> False
          ) && Maybe.isJust 
                 (List.find 
                    (\ta -> case ta of
                              TypeArgument_KindStarCase _ -> True
                              _                           -> False
                    )
                    consigs_for_name               
                 )
            -- all this just checks if the request is covered by
            -- a kind * catchall case; this should be cleaned up,
            -- i.e. it should be detectable in some easier way ...
            -- Anyway, it should _not_ be covered by a kind * case
           )
     , Request_TypeImply (other_type,implied_types) <- typeImps
            -- we check the type implications one by one
     , tp `matchType` other_type
            -- the implication applies if the types match
     , x <- implied_types
            -- if the type implication applies, |name| has to
            -- be specialised to all |implied_types|
     ]

  -- there is nothing tricky about bimap implications
  bimap_implications             :: [RawRequest] -> [RawRequest]
  bimap_implications requests    =  
    [ RawRequest (bimapName,tp)
    | RawRequest (name,_)        <- requests
    , Request_BimapImply (m,tp)  <- bimapImps
    , name == m
    ]

  fun_implications               :: [RawRequest] -> [RawRequest]
  fun_implications requests      =
    [ RawRequest (name'',reduced_type)
    | RawRequest (name,tp)       <- requests
    , Request_FunImply (name',other_requests) 
                                 <- funImps
    , name == name'
            -- now we know that the implication applies
    , (avar,RawRequest (name'',typefun))
                                 <- other_requests
            -- we take each of the implied requests ...
    , let (restvar,reduced_type) =  reduce ([avar],typefun) [tp]
            -- ... and apply the type functor therin to |tp|
    , null restvar
    ]

buildRequests                    :: Int                -- how many iterations?
                                 -> [RawRequest]       -- initial requests
                                 -> FMap Name Kind     -- generic abstractions
                                 -> [Request_TypeImply]
                                 -> [Request_BimapImply]
                                 -> [Request_FunImply]
                                 -> FMap Name [TypeArgument] -- consigs
                                 -> FMap Name Kind     -- kinds
                                 -> FMap Name KindIndexedType -- ki-types
                                 -> FMap Name LambdaType -- structure types
                                 -> Set Name           -- abstract types
                                 -> (Int               -- iterations needed
                                    ,[NormalRequest]
                                    ,[ComponentRequest]
                                    ,[KindCaseRequest]
                                    ,[GenericAbstractionRequest]
                                    ,[NormalTITRequest]
                                    ,[ComponentTITRequest]
                                    ,[GHError]
                                    )

buildRequests   n initialRequests genericAbstractions
                typeImps bimapImps funImps consigs 
                kinds kindIndexedTypes structureTypes abstractTypes =
  let consigsNames               =  mapFM (map typeArgumentName) consigs
      hasKindStar t              =  case applyFMM 
                                           kinds 
                                           (const starK)
                                           (typeName t)
                                    of
                                      Kind_Star _ -> True
                                      _           -> False
      (n',genabsR,othersR)       =  request_iterate 
                                      n initialRequests 
                                      genericAbstractions
                                      typeImps bimapImps funImps consigs
                                      kinds
      -- split off type-indexed type requests
      -- [for type-indexed type requests the name starts with an
      -- upper case letter]
      (compandnormalR,titR)      =  List.partition
                                      (\(RawRequest (n,t)) -> isVar n)
                                      othersR
      -- split off component requests
      -- [component requests do occur in the consig]
      (componentR,kindandnormalR)=  List.partition
                                      (\(RawRequest (n,t)) ->
                                          simpleType t
                                            && elem
                                                 (typeName t)
                                                 (findL consigsNames n)
                                      )
                                      compandnormalR
      -- split off kindcase requests
      -- [if they have kind star and a kind case is present]
      (kindcaseR,normalR)        =  List.partition
                                      (\(RawRequest (n,t)) ->
                                          simpleType t
                                            && hasKindStar t
                                            && elem
                                                 kindcaseName
                                                 (findL consigsNames n)
                                      )
                                      kindandnormalR
      -- split off component requests for type-indexed types
      (componenttitR,normaltitR) =  List.partition
                                      (\(RawRequest (n,t)) ->
                                          simpleType t
                                            && elem
                                                 (typeName t)
                                                 (findL consigsNames n)
                                      )
                                      titR
      -- build component requests
      componentRequests          =  [ ComponentRequest (n,tpname,kit)
                                    | RawRequest (n,tp) <- componentR
                                    , {- simpleType has been checked -}
                                      let tpname = typeName tp
                                    , let kit = maybe
                                                  (nokiterror n)
                                                  id
                                                  (find kindIndexedTypes n)
                                    ]
      -- build component requests for kind-indexed types
      componentTITRequests       =  [ ComponentTITRequest (n,tpname)
                                    | RawRequest (n,tp) <- componenttitR
                                    , {- simpleType has been checked -}
                                      let tpname = typeName tp
                                    ]
      -- build kindcase requests
      kindcaseRequests           =  [ KindCaseRequest (n,tpname,kit)
                                    | RawRequest (n,tp) <- kindcaseR
                                    , {- simpleType has been checked -}
                                      let tpname = typeName tp
                                    , let kit = maybe
                                                  (nokiterror n)
                                                  id
                                                  (find kindIndexedTypes n)
                                    ]
      -- build generic abstraction requests
      genericabstractionRequests =  [ GenericAbstractionRequest (n,tp)
                                    | RawRequest (n,tp) <- genabsR
                                    ]
      -- build normal requests
      normalRequests             =  [ NormalRequest (n,tpname,stp,kit)
                                    | RawRequest (n,tp) <- normalR
                                    , simpleType tp
                                    , let tpname = typeName tp
                                    , let stp = maybe
                                                  (nostperror tpname)
                                                  id
                                                  (find 
                                                    structureTypes 
                                                    (strName tpname)
                                                  )
                                    , let kit = maybe
                                                  (nokiterror n)
                                                  id
                                                  (find kindIndexedTypes n)
                                    ]
      -- build normal requests for type indexed types
      normalTITRequests          =  [ NormalTITRequest (n,tpname,stp)
                                    | RawRequest (n,tp) <- normaltitR
                                    , simpleType tp
                                    , let tpname = typeName tp
                                    , let stp = maybe
                                                  (nostperror tpname)
                                                  id
                                                  (find 
                                                    structureTypes 
                                                    (strName tpname)
                                                  )
                                    ]

      -- nontermination error; if maximum number of iterations has been reached
      nonTerminationError        =  if n' < 0 then
                                      [GHError_RequestNonTerminate n]
                                    else
                                      []
      -- specialisation error; if specialisation to an abstract type is
      -- requested
      specialisationErrors       =  [ GHError_RequestAbstractError name tp
                                    | RawRequest (name,tp) 
                                        <- normalR ++ normaltitR
                                    , simpleType tp
                                    , Maybe.isNothing 
                                        (find 
                                           structureTypes 
                                           (strName (typeName tp))
                                        )
                                      ||
                                      Set.elementOf
                                        (typeName tp)
                                        abstractTypes
                                    ]
      -- internal errors
      nokiterror n =
          internalError $
            "RequestUtils.buildRequests: "
              ++ "no kind-indexed type for " ++ unPretty (ppName n)
      nostperror n =
          internalError $
            "RequestUtils.buildRequests: "
              ++ "no structure type for " ++ unPretty (ppName n)

  in  (n - n'   -- iterations needed
      ,normalRequests
      ,componentRequests
      ,kindcaseRequests
      ,genericabstractionRequests
      ,normalTITRequests
      ,componentTITRequests
      ,nonTerminationError ++ specialisationErrors
      )

-- Debugging functions:
-- The followings functions can be used to semi-pretty-print requests
-- and implications on the screen.

showTypeImply                    :: Request_TypeImply -> String
showTypeImply (Request_TypeImply (tp,tps)) =
    PP.unPretty $ 
      hang 33 $
      fillBreak 30 (text "[N]:" <> PP.ppType tp)
        <+> text "=>"
        <+> vcat (map (\t -> text "[N]:" <> PP.ppType t) tps)

showBimapImply                   :: Request_BimapImply -> String
showBimapImply (Request_BimapImply (name,tp)) =
    PP.unPretty $
      hang 33 $
      fillBreak 30 (PP.ppName name <> text ":[T]")
        <+> text "=>"
        <+> (PP.ppName bimapName <> text ":" <> PP.ppType tp)

showFunImply                     :: Request_FunImply -> String
showFunImply (Request_FunImply (name,imps)) =
    PP.unPretty $
      hang 33 $
      fillBreak 30 (PP.ppName name <> text ":[T]")
        <+> text "=>"
        <+> vcat (map (\(var,(RawRequest (name',tp))) -> 
                         PP.ppName name' <> text ":(/\\"
                           <>  PP.ppName var <> text "."
                           <>  PP.ppType tp <> text ") [T]"
                      )
                      imps
                 ) 

showRawRequest                   :: RawRequest -> String
showRawRequest (RawRequest (name,tp)) =
    showNameTypeRequest name tp

showNormalRequest                :: NormalRequest -> String
showNormalRequest (NormalRequest (name,tpname,_,_)) =
    showNameNameRequest name tpname

showGenericAbstractionRequest    :: GenericAbstractionRequest -> String
showGenericAbstractionRequest (GenericAbstractionRequest (name,tp)) =
    showNameTypeRequest name tp

showComponentRequest             :: ComponentRequest -> String
showComponentRequest (ComponentRequest (name,tpname,_)) =
    showNameNameRequest name tpname

showKindCaseRequest              :: KindCaseRequest -> String
showKindCaseRequest (KindCaseRequest (name,tpname,_)) =
    showNameNameRequest name tpname

showNormalTITRequest             :: NormalTITRequest -> String
showNormalTITRequest (NormalTITRequest (name,tpname,_)) =
    showNameNameRequest name tpname

showComponentTITRequest          :: ComponentTITRequest -> String
showComponentTITRequest (ComponentTITRequest (name,tpname)) =
    showNameNameRequest name tpname    

showNameTypeRequest              :: Name -> Type -> String
showNameTypeRequest name tp =
    PP.unPretty $
      fillBreak 30 (PP.ppName name)
        <+> text ":"
        <+> PP.ppType tp

showNameNameRequest              :: Name -> Name -> String
showNameNameRequest name tpname =
    PP.unPretty $
      fillBreak 30 (PP.ppName name)
        <+> text ":"
        <+> (PP.ppName tpname)
