-- gh -- a compiler for Generic Haskell.
-- Copyright (c) 2001 -- 2005  The Generic Haskell Team. Utrecht University
-- 
-- This library is free software; you can redistribute it and/or modify
-- it under the terms of the GNU Lesser General Public License as
-- published by the Free Software Foundation; either version 2.1 of the
-- License, or (at your option) any later version.
-- 
-- This library is distributed in the hope that it will be useful, but
-- WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-- Lesser General Public License for more details.
-- 
-- You should have received a copy of the GNU Lesser General Public
-- License along with this library; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
--
-- $Id: Utils.hs 194 2005-01-12 09:16:08Z andres $
--
-- author: Andres Loeh (andres@cs.uu.nl)

{-
 - This is a collection of very general utilities that do not fit
 - anywhere else.
 -}

module Utils                    where

import IO
import System
import Char
import ST
import List

-- Just an abbreviation for the newline character.

nl                              :: String
nl                              =  "\n"

-- And an abbreviation for the lambda backslash.

lam                             :: String
lam                             =  "\\"

-- Get terminal width. We use that to print error messages
-- (which are intended for screen output). 
-- We check for the COLUMNS variable (which is set by
-- some Unix shells, and otherwise assume 75).

getTerminalWidth                :: IO Int
getTerminalWidth                =  do c_ <- try
                                              (do cols <- getEnv "COLUMNS"
                                                  return cols
                                              )
                                      case c_ of 
                                        Right c -> case reads c of
                                                     [(c,"")] -> return c
                                                     _        -> return 75
                                        Left _  -> return 75

-- "Safe head": Give an informative error message on fail.
safeHead                        :: [a] -> String -> a
safeHead [] s                   =  error s
safeHead xs _                   =  head xs

-- Construct a comment.

comment                         :: String -> [String]
comment s 
  | '\n' `elem` s               =  ["{-\n" ++ s ++ "\n-}"]   
  | otherwise                   =  ["\n","-- " ++ s]

-- This is a more sophisticated variant of the prelude getLine that can
-- deal with explicit backspace characters in the input string. This
-- should probably go to the |Utils| module.

getLineWithBackspaces           :: IO String
getLineWithBackspaces           =  do 
                                     s <- getLine 
                                     return . reverse $ remBs [] s
    where
    remBs acc []                =  acc
    remBs acc ('\b':cs)         =  remBs (drop 1 acc) cs
    remBs acc (c   :cs)         =  remBs (c:acc)      cs

-- "Increase" a string identifier. Make a to b, and z to aa.
-- Requires well-formedness. Not very efficient.

incstring                       :: String -> String
incstring s                     =  let
                                     (l:ls) = s
                                   in 
                                     incsingle l ++ ls 
  where
  incsingle 'z'                 =  "aa"
  incsingle c                   =  [ succ c ]                           

-- Transform the first character into uppercase. Fails on the
-- empty string.

fstToUpper                      :: String -> String
fstToUpper (c:cs)               =  toUpper c : cs

fstToLower                      :: String -> String
fstToLower (c:cs)               =  toLower c : cs

-- check if a given identifier is a constructor or a variable

isConid                         :: String -> Bool
isConid []                      =  False
isConid c@(x:xs)                =  isUpper x || x == ':' || c == "[]" || c == "->"
                                     || isTupleString c

isVarid                         :: String -> Bool
isVarid []                      =  False
isVarid xs                      =  not (isConid xs)

-- check if a given string is a tuple constructor

isTupleString                   :: String -> Bool
isTupleString  ('(':rest)       =  isTupleString1 rest
  where
    isTupleString1 (',':rest)   =  isTupleString1 rest
    isTupleString1 [')']        =  True
    isTupleString1 _            =  False
isTupleString  _                =  False

-- foldr01 is a combination of foldr1 with a case for the empty list

foldr01                         :: a -> (a -> a -> a) -> [a] -> a
foldr01 z _ []                  =  z
foldr01 _ f l                   =  foldr1 f l

-- product 
cross                           :: (a -> b) -> (c -> d) -> (a,c) -> (b,d)
cross f g (x,y)                 =  (f x,g y)

-- ignores two arguments
const2                          :: c -> a -> b -> c
const2 k x y                    =  k

-- Split an infinite list into two.
splitList                       :: [a] -> ([a],[a])
splitList (x:y:ys)              =  cross (x:) (y:) (splitList ys)

--------------------------
-- A simple state monad --
--------------------------

{-
newtype StateMonad s t          =  SM (s -> (s,t))

instance Monad (StateMonad s) where
  return x                      =  SM $ \s -> (s,x)
  SM m >>= f                    =  SM $ \s -> let 
					        (s1,x) = m s
                                                SM f' = f x
                                              in f' s1

getState                        :: StateMonad s s
getState                        =  SM $ \s -> (s,s)

putState                        :: s -> StateMonad s ()
putState s                      =  SM $ \_ -> (s,())

updateState                     :: (s -> s) -> StateMonad s ()
updateState f                   =  SM $ \s -> (f s,())

incrInt                         :: StateMonad Int Int
incrInt                         =  SM $ \n -> (n+1,n)

incrInt'                        :: StateMonad (Int,state) Int
incrInt'                        =  SM $ \(n,s) -> ((n+1,s),n)

runState                        :: s -> StateMonad s a -> a
runState s (SM f)               =  let (_,r) = f s in r
-}

------------------------------------
-- Substitutions and environments --
------------------------------------

-- this part should sooner or later be replaced by the FiniteMap
-- module

{-
newtype Subst v t               =  Subst [(v,t)]
                                deriving Show
type Env v t                    =  Subst v t

-- The empty substitution.

nullSubst                       :: Ord v => Subst v t
nullSubst                       =  Subst []

-- A substitution with just one element.

(+->)                           :: Ord v => v -> t -> Subst v t
v +-> t                         =  Subst [(v,t)]

-- Create a substitution from an assoc list.

mkSubst                         :: Ord v => [(v,t)] -> Subst v t
mkSubst vts                     =  Subst vts 

-- Different lookups.

lookupSubstDefF                 :: Ord v 
                                => Subst v t -> (v -> b) -> (t -> b) -> v 
                                -> b
lookupSubstDefF 
  (Subst vts) def f v           =  case lookup v vts of
                                     Nothing -> def v
                                     Just t  -> f t

lookupSubstDef                  :: Ord v 
                                => Subst v t -> t -> v -> t
lookupSubstDef s t v            =  lookupSubstDefF s (const t) id v

lookupSubst                     :: Ord v 
                                => Subst v t -> v -> Maybe t
lookupSubst s v                 =  lookupSubstDefF s (const Nothing) Just v   

-- Compose two substitutions. 
-- We need to know how to apply a substitution for this.

composeSubst                    :: Ord v 
                                => (Subst v t -> t -> t)   -- apply
                                -> Subst v t -> Subst v t 
                                -> Subst v t
composeSubst apply
             s@(Subst s1) 
               (Subst s2)       =  Subst 
                                $  [(u,apply s t) | (u,t) <- s2] ++ s1

-- Join two substitutions.
-- If there are two substitutions for the same thing, the second one
-- wins.

joinSubst                       :: Ord v
                                => Subst v t -> Subst v t
                                -> Subst v t
joinSubst                       =  composeSubst (const id)
-}

commaList :: [String] -> String
commaList [] = ""
commaList [x] = x
commaList (x:xs) = x ++ ", " ++ commaList xs

-- tuple selectors 

fst3 (x,_,_) = x
snd3 (_,x,_) = x
thd3 (_,_,x) = x
trd3 (_,_,x) = x

-- from Helium

doubleSizeOfSTArray :: a -> STArray state Int a -> ST state (STArray state Int a)
doubleSizeOfSTArray unit starray = 
   do let (lower,upper) = boundsSTArray starray
      newarray <- newSTArray (lower,2 * upper - lower) unit
      let f i = do value <- readSTArray starray i            
                   writeSTArray newarray i value
      mapM_ f [lower..upper]
      return newarray

groupAll :: Ord a => [a] -> [[a]]
groupAll = group.sort

groupAllBy :: Ord a => (a -> a -> Bool) -> [a] -> [[a]]
groupAllBy eq = groupBy eq.sort

safeLookup :: Eq a => String -> a -> [(a,b)] -> b
safeLookup err i = maybe (error err) id . lookup i
