-- gh -- a compiler for Generic Haskell.
-- Copyright (c) 2001 -- 2005  The Generic Haskell Team. Utrecht University
-- 
-- This library is free software; you can redistribute it and/or modify
-- it under the terms of the GNU Lesser General Public License as
-- published by the Free Software Foundation; either version 2.1 of the
-- License, or (at your option) any later version.
-- 
-- This library is distributed in the hope that it will be useful, but
-- WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-- Lesser General Public License for more details.
-- 
-- You should have received a copy of the GNU Lesser General Public
-- License along with this library; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
--
-- $Id: Main.hs 218 2005-01-14 11:51:51Z andres $
--
-- author: Andres Loeh (andres@cs.uu.nl)

module Main where

import Directory
import System
import FileNameUtils
import Maybe
import System.Console.GetOpt
import List
import Monad
import OSSpecific hiding (diff, touch)
import qualified OSSpecific as OS

-- Overall structure:
-- We recursively descend in a directory tree.
-- Each test has a name that can be derived from its path.
-- Each test is in a single directory with two subdirs "ref" and "out".
-- The "ref" subdir contains the test files, the "out" subdir contains
-- the work files. The directory itself can be used to store the result
-- from the last run.

testCurrentDirectory :: Flags -> IO (Int,Int)
testCurrentDirectory flags = 
  do -- get contents of current directory
     currentDir <- getCurrentDirectory
     protocol (dirs flags) $ "[entering " ++ currentDir ++ "]"
     files      <- getDirectoryContents currentDir
     files      <- return $ sort files
     -- check for option files
     g_gh_opts'  <- catch
                      (readFile ("gh_options")) (const $ return "")
     let gh_opts' = stripnewlines $ gh_opts flags ++ " " ++ g_gh_opts'
     g_hc_opts'  <- catch 
                      (readFile ("hc_options")) (const $ return "")
     let hc_opts' = stripnewlines $ hc_opts flags ++ " " ++ g_hc_opts'
     let rflags   = flags { gh_opts = gh_opts', hc_opts = hc_opts' }
     -- check for all subdirectories
     subdirs'   <- filterM doesDirectoryExist files
     let subdirs = subdirs' \\ [".",".."]
     -- if "ref" exists, then perform test here
     (htot,hfail) 
                <- if elem "ref" subdirs then
                     performTest rflags
                   else
                     return (0,0)
     -- recursively descend into all subdirs
     recresults <- mapM (testSubDirectory rflags) (subdirs \\ ["ref","out"])
     let (total,failed) = foldr (\(t,f) (t',f') -> (t+t',f+f'))
                                (htot,hfail)
                                recresults
     protocol (dirs flags) $ "[leaving " ++ currentDir ++ "]"
     return (total,failed)

testSubDirectory :: Flags -> FilePath -> IO (Int,Int)
testSubDirectory flags dir =
  do setCurrentDirectory dir
     (t,f) <- if exclude dir then return (0,0) 
                             else testCurrentDirectory flags
     setCurrentDirectory ".."
     return (t,f)
 where
  exclude dir = filename dir `elem` [".svn","CVS"]

performTest :: Flags -> IO (Int,Int)
performTest flags =
  do protocol (debug flags) $ "performing test"
     protocolBegin (debug flags) $ "creating out directory"
     shell $ mkdirCall flags ++ " " ++ "out"
     protocolBegin (debug flags) $ "removing files in out directory"
     shell $ rmCall flags ++ " " ++ "out" ++ [s] ++ "*"
     protocolSucceed (debug flags)
     protocolBegin (debug flags) $ "copying source files"
     shell $ cpCall flags ++ " " ++ "ref" ++ [s] ++ "*.ghs out"
     protocolSucceed (debug flags)
     protocolBegin (debug flags) $ "checking for tests"
     files <- getDirectoryContents "out"
     let tfiles = filter (not . null) $ map basename files
     protocol (debug flags) $ " " ++ unwords tfiles
     results <- mapM (testFile flags) tfiles
     let total  =  length tfiles
     let failed =  length . filter (==False) $ results
     return (total,failed)

protocol :: Bool -> String -> IO ()
protocol b = if b then putStrLn else const (return ())

protocolBegin :: Bool -> String -> IO ()
protocolBegin b t = if b then putStr (t ++ " ...") else return ()

protocolSucceed :: Bool -> IO ()
protocolSucceed b = protocol b " done"

protocolTest :: Flags -> String -> IO ()
protocolTest f t = if not (debug f) 
                   then putStr ("testing " ++ t ++ " ... "
                                ++ replicate (25 - length t `max` 0) ' ') 
                   else return ()

protocolTestPhase :: Flags -> String -> IO ()
protocolTestPhase f t = if not (debug f) 
                        then putStr t else return ()

testFile :: Flags -> FilePath -> IO Bool
testFile flags file =
  do protocol (debug flags) $ "testing " ++ file
     protocolTest flags $ file
     currentDir <- getCurrentDirectory
     files      <- getDirectoryContents (currentDir ++ [s] ++ "ref")
     let g_hc_options =  "hc_options" `elem` files
     let g_gh_options =  "gh_options" `elem` files
     let rfiles =  filter (\f -> basename f == file) files
     let exts   =  map fromJust . filter isJust . map extension $ rfiles
     let ghs    =  "ghs" `elem` exts
     let hs     =  "hs"  `elem` exts
     let ghi    =  "ghi" `elem` exts
     let gh_out =  "gh_out" `elem` exts
     let hc_out =  "hc_out" `elem` exts
     let out    =  "out" `elem` exts
     let hc_options =  "hc_options" `elem` exts
     let gh_options =  "gh_options" `elem` exts

     g_gh_opts'  <- catch
                      (readFile ("ref" ++ [s] ++ "gh_options")) 
                      (const $ return "")
     gh_opts'    <- catch 
                      (readFile ("ref" ++ [s] ++ file ++ ".gh_options")) 
                      (const $ return "")
     let gh_opts_ = stripnewlines $ gh_opts flags ++ " " ++ g_gh_opts' ++ " " ++ gh_opts'
     g_hc_opts'  <- catch 
                      (readFile ("ref" ++ [s] ++ "hc_options")) 
                      (const $ return "")
     hc_opts'    <- catch 
                      (readFile ("ref" ++ [s] ++ file ++ ".hc_options")) 
                      (const $ return "")
     let hc_opts_ = stripnewlines $ hc_opts flags ++ " " ++ g_hc_opts' ++ " " ++ hc_opts'

     files      <- setCurrentDirectory (currentDir ++ [s] ++ "out")
     -- Step 1: Compile source file
     protocolTestPhase flags "[gh]"

     let gh_call = ghCall flags ++ " " ++ gh_opts_
                     ++ " " ++ (file ++ ".ghs") 
                     ++ " > " ++ (file ++ ".gh_out")
     protocol (debug flags) $ "$ " ++ gh_call
     shell gh_call
     
     s1         <- -- Step 2: Compare interface files
                   if ghi    || create flags then
                     do protocolTestPhase flags "[ghi]"
                        fileCompare flags
                          (".." ++ [s] ++ "ref" ++ [s] ++ file ++ ".ghi")
                          (file ++ ".ghi")
                   else
                     do protocolTestPhase flags "     "
                        return True

     s2         <- -- Step 3: Compare haskell source files
                   if hs     || create flags then
                     do protocolTestPhase flags "[hs]"
                        fileCompare flags
                          (".." ++ [s] ++ "ref" ++ [s] ++ file ++ ".hs")
                          (file ++ ".hs")
                   else
                     do protocolTestPhase flags "    "
                        return True

     s3         <- -- Step 4: Compare GH output
                   if gh_out || create flags then
                     do protocolTestPhase flags "[gh_out]"
                        fileCompare flags
                          (".." ++ [s] ++ "ref" ++ [s] ++ file ++ ".gh_out")
                          (file ++ ".gh_out")
                   else
                     do protocolTestPhase flags "        "
                        return True

     -- Only continue if we succeed so far
     s45        <- if s1 && s2 && s3 && (hc_out || out || create flags) then do

       -- Step 5: Compile Haskell file with GHC
       protocolTestPhase flags "[hc]"

       let hc_call = hcCall flags ++ " " ++ hc_opts_ 
                       ++ " -o " ++ file ++ exe ++ " " ++ (file ++ ".hs")
                       ++ " 2> " ++ (file ++ ".hc_out")
       protocol (debug flags) $ "$ " ++ hc_call
       shell hc_call

       s4       <- -- Step 6: Compare GHC output
                   if hc_out || create flags then 
                     do protocolTestPhase flags "[hc_out]"
                        fileCompare flags 
                          (".." ++ [s] ++ "ref" ++ [s] ++ file ++ ".hc_out") 
                          (file ++ ".hc_out") 
                   else 
                     do protocolTestPhase flags "        "
                        return True

       -- Only continue if we succeed so far
       s5       <- if s4 && (out || create flags) then do

         -- Step 6: Execute compiled file
         protocolTestPhase flags "[exe]"

         let out_call = "." ++ [s] ++ file ++ exe ++ " > " ++ (file ++ ".out") ++ " 2>&1"
         protocol (debug flags) $ "$ " ++ out_call
         shell out_call

         protocolTestPhase flags "[out]"
         fileCompare flags 
            (".." ++ [s] ++ "ref" ++ [s] ++ file ++ ".out") 
            (file ++ ".out")

        else do protocolTestPhase flags "          "
                return True
       return $ s4 && s5
      else do protocolTestPhase flags "                      "
              return True

     setCurrentDirectory ".."

     let result  =  s1 && s2 && s3 && s45
     if result then
        do protocolTestPhase flags " passed\n"
           protocol (debug flags) $ "passed " ++ file ++ "\n"
      else 
       do protocolTestPhase flags " FAILED\n"
          protocol (debug flags) $ "FAILED " ++ file ++ "\n"
          let failed_call = touchCall flags ++  " " ++ file ++ ".FAILED"
          putStrLn $ "$ " ++ failed_call
          shell failed_call
          return ()

     return result

-- file1: reference, file2: current
fileCompare flags file1 file2 =
  do let diff_call = if create flags
                     then cpCall flags ++ " " ++ file2 ++ " " ++ file1
                     else diffCall flags ++ " " ++ file1 ++ " " ++ file2
     protocol (debug flags) $ "$ " ++ diff_call
     result <- shell diff_call
     case result of
       ExitSuccess   -> return True
       ExitFailure _ -> do protocol (debug flags) "FILES DO NOT MATCH!!"
                           return False

stripnewlines = map stripnewline
  where stripnewline '\n' = ' '
        stripnewline c    = c

data Flags = Flags 
               { debug    :: Bool
               , dirs     :: Bool
               , topdir   :: String
               , gh_opts  :: String
               , hc_opts  :: String
               , gh_bin   :: String
               , hc_bin   :: String
               , gh_libs  :: String
               , diff     :: String
               , touch    :: String
               , create   :: Bool
               }

main = 
  do 
     args <- getArgs
     flags <- selectOptions args
     (total,failed) <- testSubDirectory flags (topdir flags)
     putStrLn $ show total ++ " tests total"
     if failed == 0 then putStrLn "all tests passed"
                    else putStrLn $ show failed ++ " tests failed"

initialFlags = do -- try to find some settings in the environment
                  gh_home <- catch (System.getEnv "GH_HOME") 
                                   (const $ return "")
                  ghlibs  <- catch (System.getEnv "GH_LIBRARY_PATH")
                                   (const $ return 
                                              (if null gh_home then "" 
                                                 else gh_home ++ [s] ++ "lib"
                                              )
                                   )
                  ghlibs  <- if ':' `elem` ghlibs || null ghlibs
                               then return ghlibs
                               else do ps <- descendFrom ghlibs
                                       return $ concat . intersperse ":" $ ps
                  gh      <- if not (null gh_home)
                             then  do  let gh_inplace = gh_home ++ [s] ++ "bin" ++ [s] ++ "gh" ++ bat
                                       b <- doesFileExist gh_inplace
                                       if b then return ("\"" ++ gh_inplace ++ "\"") else return "gh"
                             else  return $ "gh" ++ bat
                  ghc     <- catch (System.getEnv "GHC")
                                   (const $ if not (null gh_home)
                                            then  do  let ghc_inplace = gh_home ++ [s] ++ "bin" ++ [s] ++ "gh-ghc" ++ bat
                                                      b <- doesFileExist ghc_inplace
                                                      if b then return ("\"" ++ ghc_inplace ++ "\"") else return "gh-ghc"
                                            else  return "gh-ghc")
                  return
                     (Flags { debug = False
                            , dirs = True
                            , topdir = "."
                            , gh_opts = ""
                            , hc_opts = ""
                            , gh_bin = gh
                            , hc_bin = ghc
                            , gh_libs = if null gh_home
                                          then ghlibs
                                          else "" -- not necessary in standard conf
                            , diff = OS.diff
                                       {- if operatingSystem == "Win32"
                                          then "diff -w"
                                          else "diff" -}
                            , touch = OS.touch
                            , create = False
                            }
                     )

selectOptions args =
  case getOpt Permute options args of
    (o,[],[])  ->  do flags <- initialFlags
                      return (foldr ($) flags o)
    (_,_,errs) ->  do putStrLn $ concat errs ++ usageInfo uheader options
                      exitFailure
    
ghCall f    = gh_bin f ++ if (not $ null $ gh_libs f) then " -L" ++ gh_libs f else ""
hcCall f    = hc_bin f ++ if (not $ null $ gh_libs f) then " -i" ++ gh_libs f else ""
diffCall f  = diff f
touchCall f = touch f
mkdirCall f = OS.mkdir
rmCall f    = OS.rm
cpCall f    = OS.cp

uheader  :: String
uheader  =  "Usage:\n\ttestGH [options ...]\n"

options  :: [OptDescr (Flags -> Flags)]
options  = 
  [ Option ['c'] ["create"] (NoArg (\f -> f { create = True }))
    "create test instead of performing -- use with extreme caution!"
  , Option ['d'] ["debug"]  (NoArg (\f -> f { debug = True }))
    "show debug output"
  , Option ['s'] ["silent"] (NoArg (\f -> f { debug = False, dirs = False }))
    "no directory output"
  , Option ['t'] ["topdir"] (ReqArg (\dir f -> f { topdir = dir }) "DIR")
    "start in directory DIR"
  ]

s = OS.directorySeparator
