-- gh -- a compiler for Generic Haskell.
-- Copyright (c) 2001 -- 2005  The Generic Haskell Team. Utrecht University
-- 
-- This library is free software; you can redistribute it and/or modify
-- it under the terms of the GNU Lesser General Public License as
-- published by the Free Software Foundation; either version 2.1 of the
-- License, or (at your option) any later version.
-- 
-- This library is distributed in the hope that it will be useful, but
-- WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-- Lesser General Public License for more details.
-- 
-- You should have received a copy of the GNU Lesser General Public
-- License along with this library; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
--
-- $Id: TypeUtils.hs 194 2005-01-12 09:16:08Z andres $
--
-- author: Andres Loeh (andres@cs.uu.nl)

module TypeUtils where

import AbstractSyntax
import Name
import ErrorUtils
import TypeConstruction

-- strip foralls on toplevel
haskell98Type                    :: Type -> Type
haskell98Type tp                 =
  case tp of
    Type_Forall _ _ t            -> haskell98Type t
    Type_Parenthesized _ t       -> haskell98Type t
    t                            -> t

splitLambdaType                  :: LambdaType -> [Type]
splitLambdaType                  =  splitType . snd

splitType                        :: Type -> [Type]
splitType tp                     =
  case tp of
    Type_Application _ _ f a     -> splitType f ++ concatMap splitType a
    Type_Variable _ _            -> []
    t@(Type_Constructor _ _)     -> [t]
    Type_Qualified _ _ t         -> splitType t {- or [] ?? -}
    Type_Forall _ _ t            -> splitType t {- or [] ?? -}
    Type_Exists _ _ t            -> splitType t {- or [] ?? -}
    Type_Parenthesized _ t       -> splitType t
    Type_TypeApplication _ _ _   -> [] {- or error ?? -}
    Type_KindApplication _ _ _   -> 
      unstaticError $
        "TypeUtils.splitType: kind application not implemented "
          ++ "(this should be fixed)" {- !! -}
    Type_Con r _                 -> [Type_Constructor r conName]
    Type_Label r _               -> [Type_Constructor r labelName]
    t@(Type_ConstructorCase _ _) -> [t]
    t@(Type_KindStarCase _)      -> [t]
    
simpleType                       :: Type -> Bool
simpleType tp                    =
  case tp of
    Type_Constructor _ _         -> True
    Type_ConstructorCase _ _     -> True
    _                            -> False

matchLambdaType                  :: LambdaType -> LambdaType -> Bool
matchLambdaType (v1,t1) (v2,t2)  =  v1 == v2 && matchType t1 t2

matchType                        :: Type -> Type -> Bool
matchType t1 t2                  =
  case (t1,t2) of
    (Type_Application _ _ f1 a1,Type_Application _ _ f2 a2)
      -> and $ zipWith matchType (f1:a1) (f2:a2)
    (Type_Constructor _ c1,Type_Constructor _ c2)
      -> c1 == c2
    (Type_Con r _,tp)
      -> matchType (Type_Constructor r conName) tp
    (tp,Type_Con r _)
      -> matchType tp (Type_Constructor r conName)
    (Type_Label r _,tp)
      -> matchType (Type_Constructor r labelName) tp
    (tp,Type_Label r _)
      -> matchType tp (Type_Label r labelName)
    (Type_Parenthesized _ tp1,tp2)
      -> matchType tp1 tp2
    (tp1,Type_Parenthesized _ tp2)
      -> matchType tp1 tp2
    (Type_ConstructorCase _ n1,Type_ConstructorCase _ n2)
      -> n1 == n2
    (Type_ConstructorCase r n,tp)
      -> matchType (Type_Constructor r (concaseName n)) tp
    (tp,Type_ConstructorCase r n)
      -> matchType tp (Type_Constructor r (concaseName n))
    (Type_KindStarCase _,Type_KindStarCase _)
      -> True
    _ -> False

typeArgumentToType               :: TypeArgument -> Type
typeArgumentToType tp =
  case tp of
    TypeArgument_DependencyStyle r n d vs
                                 -> Type_Constructor r n {- ?? -}
    TypeArgument_NamedType r n   -> Type_Constructor r n
    TypeArgument_Variable r n    -> Type_Variable r n
    TypeArgument_Constructor r _ -> Type_Constructor r conName
    TypeArgument_Label r _       -> Type_Constructor r labelName
    TypeArgument_ConstructorCase r n
                                 -> Type_Constructor r (concaseName n)
                             -- or: Type_ConstructorCase r n ??
    TypeArgument_KindStarCase r
                                 -> Type_Constructor r kindcaseName

typeArgumentName                                :: TypeArgument -> Name
typeArgumentName (TypeArgument_DependencyStyle _ n _ _) = n {- ?? -}
typeArgumentName (TypeArgument_NamedType _ n)   =  n
typeArgumentName (TypeArgument_Variable  _ n)   =  n
    {- should (probably) be: internalError "Name.typeArgumentName: Variable case called" -}
typeArgumentName (TypeArgument_Constructor _ _) =  conName
typeArgumentName (TypeArgument_Label       _ _) =  labelName
typeArgumentName (TypeArgument_ConstructorCase _ n)
                                                =  concaseName n
typeArgumentName (TypeArgument_KindStarCase _)  =  kindcaseName

typeName                                        :: Type -> Name
typeName (Type_Variable _ n)                    =  n
typeName (Type_Constructor _ n)                 =  n
typeName (Type_Parenthesized _ tp)              =  typeName tp
typeName (Type_Con _ _)                         =  conName
typeName (Type_Label _ _)                       =  labelName
typeName (Type_ConstructorCase _ n)             =  concaseName n
typeName (Type_KindStarCase _)                  =  kindcaseName
typeName _                                      =
  internalError "TypeUtils.typeName: called on malformed type argument"

hasDependencyItems                              :: Type -> Bool
hasDependencyItems (Type_Qualified _ ct _)      =  any isDependencyItem ct
  where
    isDependencyItem (ContextItem_DependencyItem _ _ _ _ _) = True
    isDependencyItem _                                      = False
hasDependencyItems _                            =  False
