
ToDo: Polish readInteger, readRational

> module UHA_ParseUtils where 
> 
> import AbstractSyntax
> import UHA_Range
> import UHA_ParseMonad
> import ErrorUtils
> import Name
> import Configure
> import Pretty
> 
> import Char(isDigit,isOctDigit,isHexDigit,digitToInt)
> import Ratio

Check the interface file version.
Prevents incompatible interface files from being read.

> checkInterfaceVersion :: String -> Lex ()
> checkInterfaceVersion n  
>   | n == interfaceVersion = return ()
> checkInterfaceVersion _   = parseError "Wrong interface file version; please delete and recompile."

Splits a type application into the name of the type constructor and 
the list of its arguments. (ks)
26.01.2003: Added |True| in first pattern as sanity check.

> splitTyConApp :: Type -> Lex (Name,[Type])
> splitTyConApp t = split t []
>  where
> 	split :: Type -> [Type] -> Lex (Name,[Type])
> 	split (Type_Application _ True t u) ts = split t (u ++ ts)
> 	split (Type_Constructor  _ t)       ts = return (t,ts)
> 	split _ _ = parseError "Illegal data/newtype declaration"

The function |checkContext| tests if a type can be read as a
context for a qualified type. For this, it has to be a tuple of assertions
or a single assertion. Assertions are applications of type constructors to
something.

> checkContext' :: Bool   -- allow dependency specifications
>               -> Type -> Lex ContextItems
> checkContext' deps (Type_Parenthesized _ type_)
>                                = do c <- checkAssertion deps type_
>                                     return [c]
> checkContext' deps (Type_Constructor _ n)
>   | isUnitName n               = return [] {- empty context -}
> checkContext' deps (Type_Application _ False (Type_Constructor _ n) arguments)
>   | isTupleName n              = mapM (checkAssertion deps) arguments
> checkContext' deps type_
>                                = do c <- checkAssertion deps type_
>                                     return [c]

> checkContext :: Type -> Lex ContextItems
> checkContext = checkContext' False

> checkTypeSignatureContext :: Type -> Lex ContextItems
> checkTypeSignatureContext = checkContext' True

Changed for multi-parameter type classes.

> checkAssertion :: Bool   -- allow dependency specifications
>                -> Type -> Lex ContextItem
> checkAssertion deps type_ = 
>   do a <- checkAssertion' [] type_
>      case a of 
>        Left   (name,types)    ->  return $ ContextItem_ContextItem (getRange type_) name types
>        Right  (r,name,gs,ps,asn)
>          | deps               ->  return $ ContextItem_DependencyItem r name gs ps asn
>        _                      ->  parseError "Illegal class assertion"
>  where	
>   checkAssertion' ts (Type_Parenthesized _ type_) = checkAssertion' ts type_
>   checkAssertion' ts (Type_Constructor _ name)    = return $ Left (name,ts)
>   checkAssertion' ts (Type_Application _ True f t)= checkAssertion' (t++ts) f
>   checkAssertion' ts (Type_TypeApplication r name (Type_Variable _ n))
>                                                   = do  asn <- checkAs ts
>                                                         return $ Right (r,name,[n],[],asn)
>   checkAssertion' ts (Type_TypeApplication r name (Type_TypeSignature _ gs ps))
>                                                   = do  gs' <- checkNoKind gs
>                                                         asn <- checkAs ts
>                                                         return $ Right (r,name,gs',ps,asn)
>   checkAssertion' _ _                             = parseError "Illegal class assertion"
>
>   checkAs :: [Type] -> Lex MaybeName
>   checkAs [Type_Variable _ ask,Type_Variable _ asn]
>     | ask == hsAsName unknownRange []  =  return (MaybeName_Just asn)
>   checkAs []                           =  return MaybeName_Nothing
>   checkAs _                            =  parseError "Illegal dependency item"
>
>   checkNoKind :: [KindAnnotatedName] -> Lex [Name]
>   checkNoKind =  mapM checkNoKind'
>   checkNoKind' (KindAnnotatedName_KindAnnotatedName _ n MaybeKind_Nothing) = return n
>   checkNoKind' _ = parseError "Illegal class assertion"



Checking Patterns.

We parse patterns as expressions and check for valid patterns below,
converting the expression into a pattern at the same time.

> checkPattern :: Expression -> Lex Pattern
> checkPattern e = checkPat e []
> 
> checkPatterns :: Expressions -> Lex Patterns
> checkPatterns es = mapM checkPattern es
> 
> checkPat :: Expression -> Patterns -> Lex Pattern
>
> checkPat (Expression_Literal range literal) []
>   = return (Pattern_Literal range literal)
>
> checkPat (Expression_Variable range name) []
>   | getModule name == [] = return (Pattern_Variable range name)
>
> checkPat (Expression_Constructor range_begin name) patterns 
>   = do let range_end = if null patterns then range_begin else getRange patterns
>        let range = rangeFromTo range_begin range_end
>        return (Pattern_Constructor range name patterns)
>
> checkPat (Expression_Parenthesized range expression) []
>   = do expression_ <- checkPat expression []
>        return (Pattern_Parenthesized range expression_)
>
> checkPat (Expression_NormalApplication range function arguments) patterns
>   = do patterns_ <- mapM (\ x -> checkPat x []) arguments
>        checkPat function (patterns_ ++ patterns)
>
> checkPat (Expression_InfixApplication range left operator right) []
>   = do left_            <- case left of
>                              MaybeExpression_Nothing -> patFail 
>                              MaybeExpression_Just x  -> checkPat x []
>        right_           <- case right of 
>                              MaybeExpression_Nothing -> patFail
>                              MaybeExpression_Just x  -> checkPat x []
>        (succ,operator_) <- checkPatternOperator operator
>        if    succ
>         then {- we have a + operator; so it should be a successor pattern -}
>              do name     <- checkVariable       left_
>                 literal  <- checkIntegerLiteral right_
>                 return (Pattern_Successor range name literal)
>         else return (Pattern_InfixConstructor range left_ operator_ right_)
>  where
>   patFail = parseError "Sections are not allowed in patterns."
>
> checkPat (Expression_List range expressions) []
>   = do patterns_ <- mapM (\ x -> checkPat x []) expressions
>        return (Pattern_List range patterns_)
>
> checkPat (Expression_Tuple range expressions) []
>   = do patterns_ <- mapM (\ x -> checkPat x []) expressions
>        return (Pattern_Tuple range patterns_)
>
> checkPat (Expression_RecordConstruction range name bindings) []
>   = do bindings_ <- mapM (\ x -> checkPatBinding x) bindings
>        return (Pattern_Record range name bindings_)
>
> checkPat (Expression_Negate range1 (Expression_Literal range2 literal)) []
>   = do literal_ <- checkPatternLiteral literal 
>        return (Pattern_Negate range1 literal_)
>
> checkPat (Expression_As range name expression) []
>   = do pattern_ <- checkPat expression []
>        return (Pattern_As range name pattern_)
>
> checkPat (Expression_Wildcard range) []
>   = return (Pattern_Wildcard range)
>
> checkPat (Expression_Irrefutable range expression) []
>   = do pattern_ <- checkPat expression []
>        return (Pattern_Irrefutable range pattern_)
>
> checkPat e _
>   = parseError $ "The expression is not a legal pattern: " ++ unPretty (ppExpression e) ++ "."

> checkPatBinding :: RecordExpressionBinding -> Lex RecordPatternBinding
> checkPatBinding (RecordExpressionBinding_RecordExpressionBinding range name expression) 
>   = do pattern_ <- checkPat expression [] 
>        return (RecordPatternBinding_RecordPatternBinding range name pattern_)

We check if an expression is a valid infix operator in a pattern.

> checkPatternOperator :: Expression -> Lex (Bool,Name)
> checkPatternOperator (Expression_Variable _ name@(Name_Operator _ [] "+"))
>   = return (True,name)
> checkPatternOperator (Expression_Constructor _ name)
>   = return (False,name)
> checkPatternOperator _
>   = parseError "Infix operator in pattern is neither `+' nor a constructor."

We check if a literal is valid in a negation pattern.

> checkPatternLiteral :: Literal -> Lex Literal
> checkPatternLiteral l@(Literal_Int _ _)   = return l
> checkPatternLiteral l@(Literal_Float _ _) = return l
> checkPatternLiteral _                     
>   = parseError "Only negative integer and float literals are allowed in patterns."
>
> checkVariable :: Pattern -> Lex Name
> checkVariable (Pattern_Variable _ name) = return name
> checkVariable _ = parseError "Left-hand side of n+k pattern must be a variable."
>
> checkIntegerLiteral :: Pattern -> Lex Literal
> checkIntegerLiteral (Pattern_Literal _ lit@(Literal_Int _ _)) = return lit
> checkIntegerLiteral _ = parseError "Right-hand side of n+k pattern must be an integer literal."


Check Expression Syntax.
This is to ensure that there are no pattern-syntax occurences in the expressions (ks).

> checkExpr :: Expression -> Lex Expression
> checkExpr e = case e of
>   Expression_Literal _ _        -> return e
>   Expression_Variable _ _       -> return e
>   Expression_Constructor _ _    -> return e
>   Expression_Parenthesized r e' -> check1Expr e' (Expression_Parenthesized r)
>   Expression_NormalApplication r f' es'
>                                 -> do es <- checkManyExprs es' id
>                                       f  <- checkExpr f'
>                                       return (Expression_NormalApplication r f es)
>   Expression_InfixApplication r left' op' right'
>                                 -> do left  <- checkMaybeExpr left'
>                                       op    <- checkExpr op'
>                                       right <- checkMaybeExpr right'
>                                       return (Expression_InfixApplication r left op right)
>   Expression_If r g' t' e'      -> check3Exprs g' t' e' (Expression_If r)
>   Expression_Lambda r ps' e'    -> check1Expr e' (Expression_Lambda r ps')
>   Expression_Case r e' alts'    -> do alts  <- mapM checkAlt alts'
>                                       e     <- checkExpr e'
>                                       return (Expression_Case r e alts)
>   Expression_Let r ds' e'       -> check1Expr e' (Expression_Let r ds')
>   Expression_Do r stmts'        -> do stmts <- mapM checkStmt stmts'
>                                       return (Expression_Do r stmts)
>   Expression_List r es'         -> checkManyExprs es' (Expression_List r)
>   Expression_Tuple r es'        -> checkManyExprs es' (Expression_Tuple r)
>   Expression_Comprehension r e' quals'
>                                 -> do quals <- mapM checkQual quals'
>                                       e     <- checkExpr e'
>                                       return (Expression_Comprehension r e quals)
>   Expression_Typed r e' tp'     -> check1Expr e' (\ x -> Expression_Typed r x tp')
>   Expression_RecordConstruction r n' rebs'
>                                 -> do rebs <- mapM checkField rebs'
>                                       return (Expression_RecordConstruction r n' rebs)
>   Expression_RecordUpdate r e' rebs'
>                                 -> do rebs <- mapM checkField rebs'
>                                       e    <- checkExpr e'
>                                       return (Expression_RecordUpdate r e rebs)
>   Expression_Enum r fr' th' to' -> do fr <- checkExpr fr'
>                                       th <- checkMaybeExpr th'
>                                       to <- checkMaybeExpr to'
>                                       return (Expression_Enum r fr th to)
>   Expression_Negate r e'        -> check1Expr e' (Expression_Negate r)
>   Expression_TypeApplication _ _ _
>                                 -> return e
>   _                             -> parseError "Pattern syntax in expression"

Type signature for polymorphic recursion!!

> check1Expr :: Expression -> (Expression -> a) -> Lex a
> check1Expr e f = checkExpr e >>= (return . f)
>
> check2Exprs :: Expression -> Expression
>             -> (Expression -> Expression -> a)
>             -> Lex a
> check2Exprs e1 e2 f =
>       checkExpr e1 >>= \e1 ->
>       checkExpr e2 >>= \e2 ->
>       return (f e1 e2)
> 
> check3Exprs :: Expression -> Expression -> Expression 
>             -> (Expression -> Expression -> Expression -> a) 
>             -> Lex a
> check3Exprs e1 e2 e3 f = 
> 	checkExpr e1 >>= \e1 ->
> 	checkExpr e2 >>= \e2 ->
> 	checkExpr e3 >>= \e3 ->
> 	return (f e1 e2 e3)
> 
> checkManyExprs :: Expressions -> (Expressions -> a) -> Lex a
> checkManyExprs es f =
> 	mapM checkExpr es >>= \es ->
> 	return (f es) 
>
> checkMaybeExpr :: MaybeExpression -> Lex MaybeExpression
> checkMaybeExpr MaybeExpression_Nothing  = return MaybeExpression_Nothing
> checkMaybeExpr (MaybeExpression_Just e) = checkExpr e >>=
>                                           (return . MaybeExpression_Just)
> 
> checkAlt a@(Alternative_Empty _) = return a
> checkAlt (Alternative_Alternative r p' rhs')
>                            = do rhs <- checkRHS rhs'
>                                 return (Alternative_Alternative r p' rhs)
>
> checkRHS (RightHandSide_Expression r e' wh')
>                            = check1Expr e' (\ x -> RightHandSide_Expression r x wh')
> checkRHS (RightHandSide_Guarded r es' wh')
>                            = do es <- mapM checkGuardedExpression es'
>                                 return (RightHandSide_Guarded r es wh')
>
> checkGuardedExpression (GuardedExpression_GuardedExpression r g' e')
>                            = check2Exprs g' e' (GuardedExpression_GuardedExpression r)
>
> checkQual (Qualifier_Guard r g')
>                            = check1Expr g' (Qualifier_Guard r)
> checkQual q@(Qualifier_Let r d')
>                            = return q
> checkQual (Qualifier_Generator r p' e')
>                            = check1Expr e' (Qualifier_Generator r p')
> checkQual q@(Qualifier_Empty _)
>                            = return q
>
> checkStmt (Statement_Expression r e')
>                            = check1Expr e' (Statement_Expression r)
> checkStmt s@(Statement_Let r d')
>                            = return s
> checkStmt (Statement_Generator r p' e')
>                            = check1Expr e' (Statement_Generator r p')
> checkStmt s@(Statement_Empty _)
>                            = return s
> 
> checkField (RecordExpressionBinding_RecordExpressionBinding r n' e') 
>                            = check1Expr e' (RecordExpressionBinding_RecordExpressionBinding r n')

Check (generic) type signatures.
This recovers a (generic) type signature from an expression.

> checkTypeSignature :: Expression -> Type -> Lex Declaration
> checkTypeSignature gentysig type_ =
>        do e <- checkFunLhs gentysig []
>           case e of
>             Right (LeftHandSide_Function range name [])
>               -> return $ Declaration_TypeSignature
>                             range [name] type_
>             Left (name,Type_Variable range generic,[])
>               -> return $ Declaration_GenericTypeSignature
>                             range name [KindAnnotatedName_KindAnnotatedName (getRange generic) generic MaybeKind_Nothing] [] type_
>             Left (name,Type_TypeSignature range generics parametrics,[])
>               -> do  ps <- mapM checkVar parametrics
>                      return $ Declaration_GenericTypeSignature
>                                 range name generics ps type_
>             Left _
>               -> parseError "Illegal generic type signature."
>             _ -> parseError "Illegal type signature."
>   where range = rangeFromTo (getRange gentysig) (getRange type_)


Check toplevel equation Syntax.
This distinguishes pattern from normal function bindings as well as generic bindings.

> checkValDef :: (Range,Expression, RightHandSide) -> Lex Declaration
> -- checkValDef (_,e,_) | trace ("ValDef: " ++ unPretty (ppExpression e)) False = undefined
> checkValDef (range,lefthandside_, righthandside_)
>   | isFunLhs lefthandside_ =
>       do e <- checkFunLhs lefthandside_ []
>          case e of
>            Left (f,tp,ps)
>              -> do t <- checkTyArg False tp
>                    return (Declaration_TypeIndexedValue range
>                              [TypeIndexedBinding_TypeIndexedBinding 
>                                 range f t ps righthandside_])
>  	     Right lefthandside
>              -> return (Declaration_FunctionBindings range
>                           [FunctionBinding_FunctionBinding range
>                              lefthandside righthandside_]) 
>   | otherwise =
>       do pattern <- checkPattern lefthandside_
>          return (Declaration_PatternBinding range pattern righthandside_)

Check function bindings.
Polytypic and pattern binding are disallowed in some positions.

> checkSingle :: Declaration -> Lex Declaration
> checkSingle (Declaration_TypeIndexedValue _ _) = parseError "Generic function binding disallowed here."
> checkSingle d@(Declaration_PatternBinding r p rhs) 
>                                                = do checkVarPattern p
>					              return d
>   where
>     checkVarPattern (Pattern_Variable _ _)      = return ()
>     checkVarPattern (Pattern_Parenthesized _ p) = checkVarPattern p
>     checkVarPattern _  =  parseError "Complex pattern binding disallowed here."
> checkSingle x                                  = return x


Check type arguments.
We allow a more liberal syntax here now, to adapt to dependency style.
Type arguments may have an arbitrary number of variable names as arguments.
After name resolving and kind inference, the type arguments have to
be checked again for validity.

KNOWN BUG: Con and Label are checked for here; although we should check
during name resolving if we are really referring to the Con and Label
from GHPrelude at this place. For qualified names in type patterns, this
will not work correctly.

> checkTyArg :: Bool {- true if type indexed type -}
>            -> Type -> Lex TypeArgument
> checkTyArg b (Type_Parenthesized _ t') = checkTyArg b t'
> checkTyArg b (Type_Application r _ c' vs')
>            = do let c  = stripTyParens c'
>                     vs = map stripTyParens vs'
>                 case (c,vs,b) of
>                   (Type_Constructor _ (Name_Identifier _ [] "Con"), [Type_Variable _ vn], False)
>                         -> return $ TypeArgument_Constructor r vn
>                   (Type_Constructor _ (Name_Identifier _ [] "Label"), [Type_Variable _ vn], False)
>                         -> return $ TypeArgument_Label r vn
>                   (Type_Constructor _ c@(Name_Identifier _ [] "Con"), (v:vs), b) ->
>                        do (v:vs) <- mapM checkArg (v:vs)
>                           return $ TypeArgument_DependencyStyle r c (MaybeName_Just v) vs
>                   (Type_Constructor _ c@(Name_Identifier _ [] "Label"), (v:vs), b) ->
>                        do (v:vs) <- mapM checkArg (v:vs)
>                           return $ TypeArgument_DependencyStyle r c (MaybeName_Just v) vs
>                   (Type_Constructor _ c, vs, b) ->
>                        do vs <- mapM checkArg vs
>                           return $ TypeArgument_DependencyStyle r c MaybeName_Nothing vs
>                   _     -> parseError "Invalid application in type argument."
> checkTyArg _ (Type_ConstructorCase r n)
>            = return $ TypeArgument_ConstructorCase r n
> checkTyArg _ (Type_KindStarCase r)
>            = return $ TypeArgument_KindStarCase r
> checkTyArg _ (Type_Variable r n)
>            = return $ TypeArgument_Variable r n
> checkTyArg _ (Type_Constructor r (Name_Operator _ [] ":+:"))
>            = return $ TypeArgument_NamedType r sumName
> checkTyArg _ (Type_Constructor r (Name_Operator _ [] ":*:"))
>            = return $ TypeArgument_NamedType r prodName
> checkTyArg _ (Type_Constructor r n)
>            = return $ TypeArgument_NamedType r n
> checkTyArg _ _ = parseError "Malformed type argument in generic declaration."

Remove type level parentheses.

> stripTyParens :: Type -> Type
> stripTyParens (Type_Parenthesized _ t) = stripTyParens t
> stripTyParens t                        = t



We check whether an expression can be interpreted as a left-hand-side
of a declaration.

> isFunLhs :: Expression -> Bool
> isFunLhs (Expression_InfixApplication _ _ (Expression_Variable _ _) _)     = True
> isFunLhs (Expression_NormalApplication _ (Expression_Parenthesized _ e) _) = isFunLhs e
> isFunLhs (Expression_NormalApplication _ (Expression_InfixApplication _ MaybeExpression_Nothing (Expression_Variable _ _) MaybeExpression_Nothing) _)      
>                                                                            = True
> isFunLhs (Expression_NormalApplication _ (Expression_Variable _ v) _)      = True
> isFunLhs (Expression_NormalApplication _ (Expression_TypeApplication _ _ _) _) = True
> -- isFunLhs (Expression_NormalApplication _ e _)                              = isFunLhs e

<generic-haskell>

> isFunLhs (Expression_TypeApplication _ _ _)                                = True

</generic-haskell>

> isFunLhs _                                                                 = False

We try to recover a function binding from expression
syntax at this point. Returned is a pair of the function name, and
the list of arguments, if successful.


> checkFunLhs :: Expression -> Patterns
>             -> Lex (Either (Name,Type,Patterns) LeftHandSide)
> -- checkFunLhs e _ | trace ("FunLhs: " ++ unPretty (ppExpression e)) False = undefined
> checkFunLhs (Expression_Parenthesized range_ expression) args
>   = do e <- checkFunLhs expression []
>        case e of
>          Left  (n,t,p) -> return $ Left (n,t,args ++ p)
>          Right l       -> return $ Right (LeftHandSide_Parenthesized range l args)
>  where range = if null args then range_ else rangeFromTo range_ (getRange args)
> checkFunLhs (Expression_InfixApplication range
>                MaybeExpression_Nothing 
>                (Expression_Variable _ name) 
>                MaybeExpression_Nothing) args
>   = do return $ Right (LeftHandSide_Function range name args)
> checkFunLhs (Expression_InfixApplication range 
>                (MaybeExpression_Just leftExpression) 
>                (Expression_Variable _ name) 
>                (MaybeExpression_Just rightExpression)) []
>   = do leftPattern  <- checkPattern leftExpression
>        rightPattern <- checkPattern rightExpression
>        return $ Right (LeftHandSide_Infix range leftPattern name rightPattern)
> checkFunLhs (Expression_NormalApplication range function arguments) args
>   = do patterns     <- checkPatterns arguments
>        checkFunLhs function (patterns ++ args)
> checkFunLhs (Expression_Variable range_ name) args
>   = return $ Right (LeftHandSide_Function range name args)
>  where range = if null args then range_ else rangeFromTo range_ (getRange args)

<generic-haskell>

> checkFunLhs (Expression_TypeApplication range name type_) args
>   = return $ Left (name,type_,args)

</generic-haskell>

> checkFunLhs _ _
>   = parseError "Illegal value definition."

Convert a type into an instance declaration. 

> checkInstance :: Type -> Lex (ContextItems,Name,Types)
> checkInstance (Type_Qualified _ ctx tp)
>   = do (n,tps) <- flattenType tp []
>        return (ctx,n,tps)
> checkInstance tp
>   = do (n,tps) <- flattenType tp []
>        return ([],n,tps)

> flattenType :: Type -> Types -> Lex (Name,Types)
> flattenType (Type_Parenthesized _ t)   args = flattenType t args
> flattenType (Type_Application _ _ l r) args = flattenType l (r ++ args)
> flattenType (Type_Constructor _ n)     args = return (n,args)
> flattenType _ _ = parseError "Malformed instance declaration."

Dissect a type. This is similar to the case for instances, but
we have to additionally check whether the arguments are all type
variables.

> checkSimpletype :: Type -> Lex (ContextItems,SimpleType)
> checkSimpletype tp 
>   = do (ctx,qn,tps) <- checkInstance tp
>        checkUnQual qn
>        args <- mapM checkVar tps
>        return (ctx,SimpleType_SimpleType (range qn args) qn args)
>   where range_end n args = if null args then getRange n else getRange args
>         range n args = rangeFromTo (getRange n) (range_end n args)

> checkVar :: Type -> Lex KindAnnotatedName
> checkVar (Type_Kinded r n k) = return (KindAnnotatedName_KindAnnotatedName
>                                        r n (MaybeKind_Just k)
>                                       )
> checkVar (Type_Variable r n) = return (KindAnnotatedName_KindAnnotatedName 
>                                        r n (MaybeKind_Nothing)
>                                       )
> checkVar _ = parseError "Illegal type argument; variable required"


Check for a simple context. 

> checkSimplecontext :: ContextItems -> Lex ContextItems
> checkSimplecontext its = mapM checkSimpleclass its
>   where checkSimpleclass c@(ContextItem_ContextItem r n tps)
>           = do _ <- mapM checkArg tps
>                return c
>
> checkArg :: Type -> Lex Name
> checkArg (Type_Variable _ n) = return n
> checkArg _ = parseError "Illegal type argument; variable required"

Check that an identifier or symbol is unqualified. 
For occasions when doing this in the grammar would cause conflicts.

> checkUnQual :: Name -> Lex Name
> checkUnQual n@(Name_Identifier _ [] _) = return n
> checkUnQual n@(Name_Operator   _ [] _) = return n
> checkUnQual n@(Name_Special    _ [] _) = return n
> checkUnQual _                          = parseError "Illegal qualified name."

Miscellaneous utilities.

> checkPrec :: String -> Lex String
> checkPrec cs@[c] | isDigit c = return cs
> checkPrec cs                 = parseError ("Illegal precedence " ++ cs)

Stolen from Hugs' Prelude.

> readInteger :: String -> Integer
> readInteger ('0':'o':ds) = readInteger2  8 isOctDigit ds
> readInteger ('0':'x':ds) = readInteger2 16 isHexDigit ds
> readInteger          ds  = readInteger2 10 isDigit    ds
> 
> readInteger2 :: Integer -> (Char -> Bool) -> String -> Integer
> readInteger2 radix isDig ds 
>   = foldl1 (\n d -> n * radix + d) (map (fromIntegral . digitToInt) ds)

Hack ...

> readRational :: String -> Rational
> readRational xs = (readInteger (i++m))%1 * 10^^(case e of {[] -> 0;  ('+':e2) -> read e2; _ -> read e} - length m)
>   where (i,r1) = span isDigit xs
>         (m,r2) = span isDigit (dropWhile (=='.') r1)
>         e      = dropWhile (=='e') r2
>
  
Distinguish between record constructions and updated (ks).

> mkRecConstrOrUpdate :: Range -> Expression -> RecordExpressionBindings -> Lex Expression
> mkRecConstrOrUpdate r (Expression_Constructor _ c) fs 
>                = return (Expression_RecordConstruction r c fs)
> mkRecConstrOrUpdate r exp fs@(_:_)   
>                = return (Expression_RecordUpdate r exp fs)
> mkRecConstrOrUpdate _ _ _
>                = parseError "Empty record update."

Checking kind syntax.

> checkStar :: Name -> Lex ()
> checkStar (Name_Operator _ _ "*") = return ()
> checkStar _                     = parseError "Illegal kind syntax; * expected."


Grouping declarations.

Sometime later I should think about how to rewrite this function using
higher order functions. Essentially, this is nothing more than a 
|foldr01|.

> groupDecls                    ::  Declarations -> Declarations
> groupDecls []                 =   []
> groupDecls [s]                =   [s]
> groupDecls (s1 : ss)          =   
>   let
>     (s2 : r) = groupDecls ss
>     f = case (s1,s2) of {

Kind-indexed types.
We can have multiple declarations for one polykinded type,
but they must be next to each other. We therefore compare
the name of the declared type.

>           (Declaration_KindIndexedType r1
>             k1@(KindIndexedBinding_KindIndexedBinding _ name1 _ _ _ : _)
>           ,Declaration_KindIndexedType r2
>             k2@(KindIndexedBinding_KindIndexedBinding _ name2 _ _ _ : _)
>           ) | name1 == name2 
>                -> [Declaration_KindIndexedType (rangeFromTo r1 r2) (k1 ++ k2)]

Ordinary function definitions. 
These are also grouped by name equality.

>         ; (Declaration_FunctionBindings r1
>             b1@(FunctionBinding_FunctionBinding _ lhs1 _ : _)
>           ,Declaration_FunctionBindings r2
>             b2@(FunctionBinding_FunctionBinding _ lhs2 _ : _)
>           ) | nameFromLhs lhs1 == nameFromLhs lhs2
>                -> [Declaration_FunctionBindings (rangeFromTo r1 r2) (b1 ++ b2)]

Generic function definitions, i.e. type-indexed values.
Here, we group by name and type argument (note that this requires
multiple cases for one type argument to refer to the type argument in
the same way ...). Maybe it would be better to group these
after name resolving.

>         ; (Declaration_TypeIndexedValue r1
>             t1@(TypeIndexedBinding_TypeIndexedBinding _ name1 tp1 _ _ : _)
>           ,Declaration_TypeIndexedValue r2
>             t2@(TypeIndexedBinding_TypeIndexedBinding _ name2 tp2 _ _ : _)
>           ) | name1 == name2 && tp1 `eqTypeArgument` tp2
>                -> [Declaration_TypeIndexedValue (rangeFromTo r1 r2) (t1 ++ t2)]

In all other cases, we do not join anything.

>         ; (_,_) ->  [s1,s2]
>         }
>   in
>     f ++ r

Equivalence of type arguments.

> eqTypeArgument                :: TypeArgument -> TypeArgument -> Bool
> eqTypeArgument (TypeArgument_DependencyStyle _ n1 _ _) (TypeArgument_DependencyStyle _ n2 _ _) = n1 == n2
> eqTypeArgument (TypeArgument_Constructor _ v1) (TypeArgument_Constructor _ v2) = v1 == v2
> eqTypeArgument (TypeArgument_Label _ v1) (TypeArgument_Label _ v2) = v1 == v2
> eqTypeArgument tp1 tp2 = tp1 == tp2

Grouping interface file declarations.
This is essentially a wrapper around groupDecls.

> groupIDecls                   :: InterfaceDeclarations -> InterfaceDeclarations
> groupIDecls                   =  map groupIfD . groupIDecls'
>   where
>     groupIfD (InterfaceDeclaration_Declarations r ds) = InterfaceDeclaration_Declarations r $ groupDecls ds
>     groupIfD x                   = x

> groupIDecls' []               =  []
> groupIDecls' [d]              =  [d]
> groupIDecls' (d1 : ds) =
>     let (d2 : r) = groupIDecls ds
>         f = case (d1,d2) of
>               (InterfaceDeclaration_Declarations r1 x1
>                ,InterfaceDeclaration_Declarations r2 x2) 
>                  -> [InterfaceDeclaration_Declarations (rangeFromTo r1 r2) (x1 ++ x2)]
>               (_,_) -> [d1,d2]
>     in  f ++ r
